import unittest
from tempfile import TemporaryDirectory
import csv
from math import nan

import numpy as np
from numpy.testing import assert_array_equal
import h5py

from scripts.tsv_to_mrmatrix import coarsen, parse, get_height, get_width


class CoarsenTest(unittest.TestCase):
    def test_5_layer_pyramid(self):
        tile_size = 4
        max_zoom = 4
        max_width = tile_size * 2 ** max_zoom

        with TemporaryDirectory() as tmp_dir:
            hdf5 = h5py.File(tmp_dir + "/temp.hdf5", "w")
            g = hdf5.create_group("resolutions")
            g1 = g.create_group("1")
            ds = g1.create_dataset(
                "values",
                (max_width, max_width),
                dtype="f4",
                compression="lzf",
                fillvalue=np.nan,
            )
            for y in range(max_width):
                a = np.array([float(x) for x in range(max_width)])
                ds[y, :max_width] = a

            # before coarsen()
            self.assertEqual(list(hdf5.keys()), ["resolutions"])
            self.assertEqual(list(hdf5["resolutions"].keys()), ["1"])
            self.assertEqual(list(hdf5["resolutions"]["1"].keys()), ["values"])
            self.assertEqual(list(hdf5["resolutions"]["1"]["values"].shape), [64, 64])
            self.assertEqual(
                hdf5["resolutions"]["1"]["values"][:].tolist()[0],
                [float(x) for x in range(64)],
            )

            coarsen(hdf5, type="values", tile_size=tile_size)

            # after coarsen()
            self.assertEqual(list(hdf5.keys()), ["resolutions"])
            self.assertEqual(
                list(hdf5["resolutions"].keys()), ["1", "16", "2", "4", "8"]
            )
            self.assertEqual(list(hdf5["resolutions"]["16"].keys()), ["values"])
            shapes = {"1": 64, "2": 32, "4": 16, "8": 8, "16": 4}
            for (k, v) in shapes.items():
                self.assertEqual(hdf5["resolutions"][k]["values"].shape, (v, v))
            row = [1920, 6016, 10112, 14208]
            self.assertEqual(
                hdf5["resolutions"]["16"]["values"][:].tolist(), [row, row, row, row]
            )

    def test_math(self):
        tile_size = 2
        max_zoom = 2
        max_width = tile_size * 2 ** max_zoom

        with TemporaryDirectory() as tmp_dir:
            hdf5 = h5py.File(tmp_dir + "/temp.hdf5", "w")
            g = hdf5.create_group("resolutions")
            g1 = g.create_group("1")
            ds = g1.create_dataset(
                "values",
                (max_width, max_width),
                dtype="f4",
                compression="lzf",
                fillvalue=np.nan,
            )
            for y in range(max_width):
                a = np.array([float(x) for x in range(max_width)])
                ds[y, :max_width] = a

            coarsen(hdf5, type="values", tile_size=tile_size)

            # after coarsen()
            self.assertEqual(list(hdf5.keys()), ["resolutions"])
            self.assertEqual(list(hdf5["resolutions"].keys()), ["1", "2", "4"])

            shapes = {"1": 8, "2": 4, "4": 2}
            for (k, v) in shapes.items():
                self.assertEqual(hdf5["resolutions"][k]["values"].shape, (v, v))

            row8 = list(range(8))
            assert_array_equal(
                hdf5["resolutions"]["1"]["values"], [row8 for _ in range(8)]
            )

            row4 = [8 * x + 2 for x in range(4)]
            assert_array_equal(
                hdf5["resolutions"]["2"]["values"], [row4 for _ in range(4)]
            )

            row2 = [24, 88]
            assert_array_equal(
                hdf5["resolutions"]["4"]["values"], [row2 for _ in range(2)]
            )


class ParseTest(unittest.TestCase):
    def test_wide_labelled_square(self):
        with TemporaryDirectory() as tmp_dir:
            csv_path = tmp_dir + "/tmp.csv"
            with open(csv_path, "w", newline="") as csv_file:
                writer = csv.writer(csv_file, delimiter="\t")
                # header:
                col_labels = ["col-{}".format(x) for x in range(513)]
                writer.writerow(col_labels)
                # body:
                for y in range(0, 3):
                    writer.writerow(["row-{}".format(y)] + [0] * 512)
                for y in range(3, 6):
                    writer.writerow(["row-{}".format(y)] + [1] * 512)
                for y in range(6, 9):
                    writer.writerow(["row-{}".format(y)] + [1, -1] * 256)
            csv_handle = open(csv_path, "r")

            hdf5_path = tmp_dir + "tmp.hdf5"
            hdf5_write_handle = h5py.File(hdf5_path, "w")

            height = get_height(csv_path)
            width = get_width(csv_path, is_labelled=True)
            parse(
                csv_handle,
                hdf5_write_handle,
                height,
                width,
                delimiter="\t",
                first_n=None,
                is_labelled=True,
            )

            hdf5 = h5py.File(hdf5_path, "r")
            self.assertEqual(
                list(hdf5.keys()), ["col_labels", "resolutions", "row_labels"]
            )
            self.assertEqual(list(hdf5["col_labels"]), col_labels[1:])
            self.assertEqual(
                list(hdf5["row_labels"]), ["row-{}".format(r) for r in range(9)]
            )

            self.assertEqual(
                list(hdf5["resolutions"].keys()), ["1", "2", "2-nan_values"]
            )

            self.assertEqual(
                list(hdf5["resolutions"]["1"].keys()), ["nan_values", "values"]
            )
            assert_array_equal(
                hdf5["resolutions"]["1"]["nan_values"], [[0] * 512] * 512
            )
            res_1 = hdf5["resolutions"]["1"]["values"]
            assert_array_equal(res_1[0], [0] * 512)
            assert_array_equal(res_1[3], [1] * 512)
            assert_array_equal(res_1[6], [1, -1] * 256)
            assert_array_equal(res_1[9], [nan] * 512)

            self.assertEqual(list(hdf5["resolutions"]["2"].keys()), ["values"])
            # TODO: We are missing nan_values at higher aggregations: Bug?
            # https://github.com/higlass/clodius/issues/62
            res_2 = hdf5["resolutions"]["2"]["values"]
            assert_array_equal(res_2[0], [0] * 256)
            # Stradles the 0 and 1 rows
            assert_array_equal(res_2[1], [2] * 256)
            assert_array_equal(res_2[2], [4] * 256)
            assert_array_equal(res_2[3], [0] * 256)  # -1 and +1 cancel out
            assert_array_equal(res_2[4], [0] * 256)
            assert_array_equal(res_2[5], [0] * 256)
            assert_array_equal(res_2[6], [0] * 256)
            # TODO: We lose nan at higher aggregations.
            # https://github.com/higlass/clodius/issues/62

    def _assert_unlabelled_roundtrip_lt_256(self, matrix, delimiter):
        with TemporaryDirectory() as tmp_dir:
            csv_path = tmp_dir + "/tmp.csv"
            with open(csv_path, "w", newline="") as csv_file:
                writer = csv.writer(csv_file, delimiter=delimiter)
                # body:
                for row in matrix:
                    writer.writerow(row)

            csv_handle = open(csv_path, "r")
            hdf5_path = tmp_dir + "tmp.hdf5"
            hdf5_write_handle = h5py.File(hdf5_path, "w")

            is_labelled = False
            height = get_height(csv_path, is_labelled=is_labelled)
            width = get_width(csv_path, is_labelled=is_labelled)
            parse(
                csv_handle,
                hdf5_write_handle,
                height,
                width,
                first_n=None,
                is_labelled=is_labelled,
                delimiter=delimiter,
            )

            hdf5 = h5py.File(hdf5_path, "r")
            self.assertEqual(list(hdf5.keys()), ["resolutions"])
            self.assertEqual(list(hdf5["resolutions"].keys()), ["1"])
            self.assertEqual(
                list(hdf5["resolutions"]["1"].keys()), ["nan_values", "values"]
            )
            assert_array_equal(
                hdf5["resolutions"]["1"]["nan_values"],
                [[0] * len(matrix[0])] * len(matrix),
            )
            assert_array_equal(hdf5["resolutions"]["1"]["values"], matrix)

    def test_unlabelled_csv(self):
        self._assert_unlabelled_roundtrip_lt_256(
            matrix=[[x + y for x in range(4)] for y in range(4)], delimiter=","
        )

    def _assert_unlabelled_roundtrip_1024(
        self, matrix, first_row=None, first_col=None, first_n=None
    ):
        delimiter = "\t"
        with TemporaryDirectory() as tmp_dir:
            csv_path = tmp_dir + "/tmp.csv"
            with open(csv_path, "w", newline="") as csv_file:
                writer = csv.writer(csv_file, delimiter=delimiter)
                # body:
                for row in matrix:
                    writer.writerow(row)

            csv_handle = open(csv_path, "r")
            hdf5_path = tmp_dir + "tmp.hdf5"
            hdf5_write_handle = h5py.File(hdf5_path, "w")

            is_labelled = False
            height = get_height(csv_path, is_labelled=is_labelled)
            width = get_width(csv_path, is_labelled=is_labelled)
            parse(
                csv_handle,
                hdf5_write_handle,
                height,
                width,
                first_n=first_n,
                is_labelled=is_labelled,
                delimiter=delimiter,
            )

            hdf5 = h5py.File(hdf5_path, "r")
            self.assertEqual(list(hdf5.keys()), ["resolutions"])
            self.assertEqual(
                list(hdf5["resolutions"].keys()),
                ["1", "2", "2-nan_values", "4", "4-nan_values"],
            )
            self.assertEqual(
                list(hdf5["resolutions"]["1"].keys()), ["nan_values", "values"]
            )
            self.assertEqual(list(hdf5["resolutions"]["4"].keys()), ["values"])
            res_4 = hdf5["resolutions"]["4"]["values"]
            if first_row:
                assert_array_equal(res_4[0], first_row)
            if first_col:
                assert_array_equal(
                    [res_4[y][0] for y in range(len(first_col))], first_col
                )

    def test_unlabelled_tsv_tall(self):
        self._assert_unlabelled_roundtrip_1024(
            matrix=[[1 for x in range(4)] for y in range(1000)],
            first_col=[16] * 250 + [0] * 6,
        )

    def test_unlabelled_tsv_wide(self):
        self._assert_unlabelled_roundtrip_1024(
            matrix=[[1 for x in range(1000)] for y in range(4)],
            first_row=[16] * 250 + [0] * 6,
        )

    def test_unlabelled_tsv_tall_first_n(self):
        self._assert_unlabelled_roundtrip_1024(
            matrix=[[1 for x in range(4)] for y in range(1000)],
            first_col=[8] + [0] * 255,
            first_n=2,
        )

    def test_unlabelled_tsv_wide_first_n(self):
        self._assert_unlabelled_roundtrip_1024(
            matrix=[[1 for x in range(1000)] for y in range(4)],
            first_row=[8] * 250 + [0] * 6,
            first_n=2,
        )



import os.path as op


def infer_filetype(filename):
    _, ext = op.splitext(filename)

    if ext.lower() == ".bw" or ext.lower() == ".bigwig":
        return "bigwig"
    elif ext.lower() == ".mcool" or ext.lower() == ".cool":
        return "cooler"
    elif ext.lower() == ".htime":
        return "time-interval-json"
    elif ext.lower() == ".hitile":
        return "hitile"

    return None


def infer_datatype(filetype):
    if filetype == "cooler":
        return "matrix"
    if filetype == "bigwig":
        return "vector"
    if filetype == "time-interval-json":
        return "time-interval"
    if filetype == "hitile":
        return "vector"



import bbi
import clodius.tiles.format as hgfo
import functools as ft
import logging
import numpy as np
import pandas as pd
import re

from concurrent.futures import ThreadPoolExecutor

MAX_THREADS = 4
TILE_SIZE = 1024

logger = logging.getLogger(__name__)

aggregation_modes = {}
aggregation_modes["mean"] = {"name": "Mean", "value": "mean"}
aggregation_modes["min"] = {"name": "Min", "value": "min"}
aggregation_modes["max"] = {"name": "Max", "value": "max"}
aggregation_modes["std"] = {"name": "Standard Deviation", "value": "std"}

range_modes = {}
range_modes["minMax"] = {"name": "Min-Max", "value": "minMax"}
range_modes["whisker"] = {"name": "Whisker", "value": "whisker"}


def get_quadtree_depth(chromsizes):
    tile_size_bp = TILE_SIZE
    min_tile_cover = np.ceil(sum(chromsizes) / tile_size_bp)
    return int(np.ceil(np.log2(min_tile_cover)))


def get_zoom_resolutions(chromsizes):
    return [2 ** x for x in range(get_quadtree_depth(chromsizes) + 1)][::-1]


def natsort_key(s, _NS_REGEX=re.compile(r"(\d+)", re.U)):
    return tuple([int(x) if x.isdigit() else x for x in _NS_REGEX.split(s) if x])


def natcmp(x, y):
    if x.find("_") >= 0:
        x_parts = x.split("_")
        if y.find("_") >= 0:
            # chr_1 vs chr_2
            y_parts = y.split("_")

            return natcmp(x_parts[1], y_parts[1])
        else:
            # chr_1 vs chr1
            # chr1 comes first
            return 1
    if y.find("_") >= 0:
        # chr1 vs chr_1
        # y comes second
        return -1

    _NS_REGEX = re.compile(r"(\d+)", re.U)
    x_parts = tuple([int(a) if a.isdigit() else a for a in _NS_REGEX.split(x) if a])
    y_parts = tuple([int(a) if a.isdigit() else a for a in _NS_REGEX.split(y) if a])

    # order of these parameters is purposefully reverse how they should be
    # ordered
    for key in ["m", "y", "x"]:
        if key in y.lower():
            return -1
        if key in x.lower():
            return 1

    try:
        if x_parts < y_parts:
            return -1
        elif y_parts > x_parts:
            return 1
        else:
            return 0
    except TypeError:
        return 1


def natsorted(iterable):
    return sorted(iterable, key=ft.cmp_to_key(natcmp))


def get_chromsizes(bwpath):
    """
    TODO: replace this with negspy

    Also, return NaNs from any missing chromosomes in bbi.fetch

    """
    chromsizes = bbi.chromsizes(bwpath)
    chromosomes = natsorted(chromsizes.keys())
    chrom_series = pd.Series(chromsizes)[chromosomes]
    return chrom_series


def abs2genomic(chromsizes, start_pos, end_pos):
    abs_chrom_offsets = np.r_[0, np.cumsum(chromsizes.values)]
    cid_lo, cid_hi = (
        np.searchsorted(abs_chrom_offsets, [start_pos, end_pos], side="right") - 1
    )
    rel_pos_lo = start_pos - abs_chrom_offsets[cid_lo]
    rel_pos_hi = end_pos - abs_chrom_offsets[cid_hi]
    start = rel_pos_lo
    for cid in range(cid_lo, cid_hi):
        yield cid, start, chromsizes[cid]
        start = 0
    yield cid_hi, start, rel_pos_hi


def tileset_info(bwpath, chromsizes=None):
    """
    Get the tileset info for a bigWig file

    Parameters
    ----------
    bwpath: string
        The path to the bigwig file from which to retrieve data
    chromsizes: [[chrom, size],...]
        A list of chromosome sizes associated with this tileset.
        Typically passed in to specify in what order data from
        the bigwig should be returned.

    Returns
    -------
    tileset_info: {'min_pos': [],
                    'max_pos': [],
                    'tile_size': 1024,
                    'max_zoom': 7
                    }
    """
    TILE_SIZE = 1024

    if chromsizes is None:
        chromsizes = get_chromsizes(bwpath)
        chromsizes_list = []

        for chrom, size in chromsizes.iteritems():
            chromsizes_list += [[chrom, int(size)]]
    else:
        chromsizes_list = chromsizes

    min_tile_cover = np.ceil(sum([int(c[1]) for c in chromsizes_list]) / TILE_SIZE)
    max_zoom = int(np.ceil(np.log2(min_tile_cover)))

    tileset_info = {
        "min_pos": [0],
        "max_pos": [TILE_SIZE * 2 ** max_zoom],
        "max_width": TILE_SIZE * 2 ** max_zoom,
        "tile_size": TILE_SIZE,
        "max_zoom": max_zoom,
        "chromsizes": chromsizes_list,
        "aggregation_modes": aggregation_modes,
        "range_modes": range_modes,
    }
    return tileset_info


def fetch_data(a):
    (bwpath, binsize, chromsizes, aggregation_mode, range_mode, cid, start, end) = a
    n_bins = int(np.ceil((end - start) / binsize))
    n_dim = 1

    if range_mode == "minMax":
        n_dim = 2

    if range_mode == "whisker":
        n_dim = 4

    x = np.zeros((n_bins, n_dim)) if n_dim > 1 else np.zeros(n_bins)

    try:
        chrom = chromsizes.index[cid]
        clen = chromsizes.values[cid]

        args = [bwpath, chrom, start, end]
        kwargs = {"bins": n_bins, "missing": np.nan}

        if range_mode == "minMax":
            x[:, 0] = bbi.fetch(*args, **dict(kwargs, summary="min"))
            x[:, 1] = bbi.fetch(*args, **dict(kwargs, summary="max"))

        elif range_mode == "whisker":
            x[:, 0] = bbi.fetch(*args, **dict(kwargs, summary="min"))
            x[:, 1] = bbi.fetch(*args, **dict(kwargs, summary="max"))
            x[:, 2] = bbi.fetch(*args, **dict(kwargs, summary="mean"))
            x[:, 3] = bbi.fetch(*args, **dict(kwargs, summary="std"))

        else:
            x[:] = bbi.fetch(*args, **dict(kwargs, summary=aggregation_mode))

        # drop the very last bin if it is smaller than the binsize
        if end == clen and clen % binsize != 0:
            x = x[:-1]
    except IndexError:
        # beyond the range of the available chromosomes
        # probably means we've requested a range of absolute
        # coordinates that stretch beyond the end of the genome
        x[:] = np.nan
    except KeyError:
        # probably requested a chromosome that doesn't exist (e.g. chrM)
        x[:] = np.nan

    return x


def get_bigwig_tile(
    bwpath,
    zoom_level,
    start_pos,
    end_pos,
    chromsizes=None,
    aggregation_mode="mean",
    range_mode=None,
):
    if chromsizes is None:
        chromsizes = get_chromsizes(bwpath)

    resolutions = get_zoom_resolutions(chromsizes)
    binsize = resolutions[zoom_level]

    cids_starts_ends = list(abs2genomic(chromsizes, start_pos, end_pos))
    with ThreadPoolExecutor(max_workers=16) as e:
        arrays = list(
            e.map(
                fetch_data,
                [
                    tuple(
                        [bwpath, binsize, chromsizes, aggregation_mode, range_mode]
                        + list(c)
                    )
                    for c in cids_starts_ends
                ],
            )
        )

    return np.concatenate(arrays)


def tiles(bwpath, tile_ids, chromsizes_map={}, chromsizes=None):
    """
    Generate tiles from a bigwig file.

    Parameters
    ----------
    tileset: tilesets.models.Tileset object
        The tileset that the tile ids should be retrieved from
    tile_ids: [str,...]
        A list of tile_ids (e.g. xyx.0.0) identifying the tiles
        to be retrieved
    chromsizes_map: {uid: []}
        A set of chromsizes listings corresponding to the parameters of the
        tile_ids. To be used if a chromsizes id is passed in with the tile id
        with the `|cos:id` tag in the tile id
    chromsizes: [[chrom, size],...]
        A 2d array containing chromosome names and sizes. Overrides the
        chromsizes in chromsizes_map

    Returns
    -------
    tile_list: [(tile_id, tile_data),...]
        A list of tile_id, tile_data tuples
    """
    TILE_SIZE = 1024
    generated_tiles = []
    for tile_id in tile_ids:
        tile_option_parts = tile_id.split("|")[1:]
        tile_no_options = tile_id.split("|")[0]
        tile_id_parts = tile_no_options.split(".")
        tile_position = list(map(int, tile_id_parts[1:3]))
        return_value = tile_id_parts[3] if len(tile_id_parts) > 3 else "mean"

        aggregation_mode = return_value if return_value in aggregation_modes else "mean"
        range_mode = return_value if return_value in range_modes else None

        tile_options = dict([o.split(":") for o in tile_option_parts])

        if chromsizes:
            chromnames = [c[0] for c in chromsizes]
            chromlengths = [int(c[1]) for c in chromsizes]
            chromsizes_to_use = pd.Series(chromlengths, index=chromnames)
        else:
            chromsizes_id = None
            if "cos" in tile_options:
                chromsizes_id = tile_options["cos"]
            if chromsizes_id in chromsizes_map:
                chromsizes_to_use = chromsizes_map[chromsizes_id]
            else:
                chromsizes_to_use = None

        zoom_level = tile_position[0]
        tile_pos = tile_position[1]

        # this doesn't combine multiple consequetive ids, which
        # would speed things up
        if chromsizes_to_use is None:
            chromsizes_to_use = get_chromsizes(bwpath)

        max_depth = get_quadtree_depth(chromsizes_to_use)
        tile_size = TILE_SIZE * 2 ** (max_depth - zoom_level)
        start_pos = tile_pos * tile_size
        end_pos = start_pos + tile_size
        dense = get_bigwig_tile(
            bwpath,
            zoom_level,
            start_pos,
            end_pos,
            chromsizes_to_use,
            aggregation_mode=aggregation_mode,
            range_mode=range_mode,
        )

        tile_value = hgfo.format_dense_tile(dense)

        generated_tiles += [(tile_id, tile_value)]
    return generated_tiles


def chromsizes(filename):
    """
    Get a list of chromosome sizes from this [presumably] bigwig
    file.

    Parameters:
    -----------
    filename: string
        The filename of the bigwig file

    Returns
    -------
    chromsizes: [(name:string, size:int), ...]
        An ordered list of chromosome names and sizes
    """
    try:
        chrom_series = get_chromsizes(filename)
        data = []
        for chrom, size in chrom_series.iteritems():
            data.append([chrom, size])
        return data
    except Exception as ex:
        logger.error(ex)
        raise Exception("Error loading chromsizes from bigwig file: {}".format(ex))



def tileset_info(filename):
    """

    Return the bounds of this tileset. The bounds should encompass the entire
    width of this dataset.

    So how do we know what those are if we don't know chromsizes? We can assume
    that the file is enormous (e.g. has a width of 4 trillion) and rely on the
    browser to pass in a set of chromsizes
    """

    pass



from __future__ import print_function

import clodius.db_tiles as cdt
import clodius.hdf_tiles as cht
import click.testing as clt
import clodius.cli.aggregate as cca
import h5py
import negspy.coordinates as nc
import numpy as np
import os.path as op
import sys

sys.path.append("scripts")

testdir = op.realpath(op.dirname(__file__))


def test_clodius_aggregate_bedfile():
    input_file = op.join(
        testdir, "sample_data", "geneAnnotationsExonsUnions.hg19.short.bed"
    )
    output_file = "/tmp/geneAnnotationsExonsUnions.hg19.short.bed"

    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedfile,
        [
            input_file,
            "--max-per-tile",
            20,
            "--importance-column",
            5,
            "--assembly",
            "hg19",
            "--output-file",
            output_file,
            "--delimiter",
            "\t",
        ],
    )

    a, b, tb = result.exc_info
    """
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    """
    assert result.exit_code == 0

    results = cdt.get_tiles(output_file, 6, 3, num_tiles=1)
    # print("results:", results)

    assert len(results[3][0]["fields"]) == 14

    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedfile,
        [
            input_file,
            "--max-per-tile",
            20,
            "--importance-column",
            5,
            "--assembly",
            "hg19",
            "--output-file",
            output_file,
        ],
    )

    assert result.exit_code == 0

    results = cdt.get_tiles(output_file, 6, 3, num_tiles=3)

    assert len(results[3][0]["fields"]) == 17


testdir = op.realpath(op.dirname(__file__))


def test_clodius_aggregate_bedgraph():
    input_file = op.join(testdir, "sample_data", "cnvs_hw.tsv")
    assembly_file = op.join(testdir, "sample_data", "test_cnvs_assembly")
    output_file = "/tmp/cnvs_hw.hitile"

    # run once to make sure it doesn't crash on a smaller genome
    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedgraph,
        [
            input_file,
            "--output-file",
            output_file,
            # '--assembly', 'grch37',
            "--chromsizes-filename",
            assembly_file,
            "--chromosome-col",
            "2",
            "--from-pos-col",
            "3",
            "--to-pos-col",
            "4",
            "--value-col",
            "5",
            "--has-header",
            "--nan-value",
            "NA",
        ],
    )

    # run again with the proper assembly
    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedgraph,
        [
            input_file,
            "--output-file",
            output_file,
            "--assembly",
            "grch37",
            # '--chromsizes-filename', assembly_file,
            "--chromosome-col",
            "2",
            "--from-pos-col",
            "3",
            "--to-pos-col",
            "4",
            "--value-col",
            "5",
            "--has-header",
            "--nan-value",
            "NA",
        ],
    )

    """
    import traceback
    a,b,tb = result.exc_info
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    """

    assert result.exit_code == 0
    f = h5py.File(output_file, "r")
    # print("tile_0_0", d)

    # print("tile:", cht.get_data(f, 22, 0))
    # return
    d = cht.get_data(f, 0, 0)

    assert not np.isnan(d[0])
    assert np.isnan(d[-1])
    cht.get_data(f, 3, 0)
    # TODO: Make assertions about result

    # print("prev_tile_3_0:", prev_tile_3_0)

    assert result.exit_code == 0

    # TODO: Why are we ignoring these?
    # assert(sum(prev_tile_3_0) < 0)
    # input_file = op.join(testdir, 'sample_data', 'cnvs_hw.tsv.gz')
    # result = runner.invoke(
    # cca.bedgraph,
    # [input_file,
    # '--output-file', output_file,
    # '--assembly', 'grch37',
    # '--chromosome-col', '2',
    # '--from-pos-col', '3',
    # '--to-pos-col', '4',
    # '--value-col', '5',
    # '--has-header',
    # '--nan-value', 'NA'])
    # '''
    # import traceback
    # print("exc_info:", result.exc_info)
    # a,b,tb = result.exc_info
    # print("result:", result)
    # print("result.output", result.output)
    # print("result.error", traceback.print_tb(tb))
    # print("Exception:", a,b)
    # '''
    # f = h5py.File(output_file)
    # tile_3_0 = cht.get_data(f, 3, 0)
    # assert(sum(tile_3_0) - sum(prev_tile_3_0) < 0.0001)


testdir = op.realpath(op.dirname(__file__))


def test_clodius_aggregate_bedpe():
    input_file = op.join(testdir, "sample_data", "Rao_RepA_GM12878_Arrowhead.txt")
    output_file = "/tmp/bedpe.db"

    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedpe,
        [
            input_file,
            "--output-file",
            output_file,
            "--assembly",
            "hg19",
            "--chr1-col",
            "1",
            "--from1-col",
            "2",
            "--to1-col",
            "3",
            "--chr2-col",
            "1",
            "--from2-col",
            "2",
            "--to2-col",
            "3",
        ],
    )

    """
    print("result:", result)
    print("result.output", result.output)
    """

    assert result.exit_code == 0

    tiles = cdt.get_2d_tiles(output_file, 0, 0, 0, numx=1, numy=1)

    assert "\n" not in tiles[(0, 0)][0]["fields"][2]


testdir = op.realpath(op.dirname(__file__))


def test_clodius_aggregate_bedgraph1():
    input_file = op.join(testdir, "sample_data", "dm3_values.tsv")
    output_file = "/tmp/dm3_values.hitile"

    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedgraph, [input_file, "--output-file", output_file, "--assembly", "dm3"]
    )

    a, b, tb = result.exc_info

    """
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    """

    # print("result.output", result.output)

    f = h5py.File("/tmp/dm3_values.hitile", "r")
    # max_zoom = f['meta'].attrs['max-zoom']
    # TODO: Make assertions about result
    values = f["values_0"]

    import numpy as np

    # print("values:", values[8])
    # genome positions are 0 based as stored in hitile files
    assert np.isnan(values[8])
    assert values[9] == 1
    assert values[10] == 1
    assert values[13] == 1
    assert np.isnan(values[14])
    assert np.isnan(values[15])

    chrom_info = nc.get_chrominfo("dm3")
    chr_2r_pos = nc.chr_pos_to_genome_pos("chr2R", 0, chrom_info)
    # print('chr_2r_pos:', chr_2r_pos)

    assert np.isnan(values[chr_2r_pos + 28])
    assert values[chr_2r_pos + 29] == 77
    assert values[chr_2r_pos + 38] == 77
    assert values[chr_2r_pos + 39] == 0

    assert result.exit_code == 0

    d = cht.get_data(f, 0, 0)
    # print("d[:10]", d[:10])
    # print("sum(d):", sum([x for x in d if not np.isnan(x)]))
    assert np.nansum(d) > 1.0 and np.nansum(d) < 10.0

    return

    input_file = op.join(testdir, "sample_data", "test3chroms_values.tsv")
    output_file = "/tmp/test3chroms_values.hitile"

    runner = clt.CliRunner()
    result = runner.invoke(
        cca.bedgraph,
        [input_file, "--output-file", output_file, "--assembly", "test3chroms"],
    )

    # print('output:', result.output, result)

    f = h5py.File("/tmp/test3chroms_values.hitile")
    # f['meta'].attrs['max-zoom']
    # TODO: Make assertions about result

    # print('max_zoom:', max_zoom)
    # print("len", len(f['values_0']))

    values = f["values_0"]

    # print('values', values[:100])

    # genome positions are 0 based as stored in hitile files
    assert values[8] == 0
    assert values[9] == 1
    assert values[10] == 1
    assert values[13] == 1
    assert values[14] == 0
    assert values[15] == 0

    chr2_pos = nc.chr_pos_to_genome_pos("chr2", 0, "test3chroms")

    assert values[chr2_pos + 28] == 0
    assert values[chr2_pos + 29] == 77
    assert values[chr2_pos + 38] == 77
    assert values[chr2_pos + 39] == 0

    assert result.exit_code == 0

    d = cht.get_data(f, 0, 0)
    assert sum(d) == 770 + 880 + 5
    # print("d:", d)


testdir = op.realpath(op.dirname(__file__))



import clodius.tiles.utils as hgut
import json


def tileset_info(df, x_column, y_column):
    minx = df[x_column].min()
    maxx = df[x_column].max()

    miny = df[y_column].min()
    maxy = df[y_column].max()

    return {
        "min_pos": [minx, miny],
        "max_pos": [maxx, maxy],
        "max_width": max(maxy - miny, maxx - minx),
        "max_zoom": 100,
    }


def format_data(tiles):
    new_out = []

    for tile_id, tile_data in tiles:
        # print("len:", len(tile_data))
        # tile_data['data'] = tile_data['gene1']
        new_out += [(tile_id, json.loads(tile_data.to_json(orient="records")))]

    return new_out


def tiles(df, x_column, y_column, tileset_info, z, x, y, width=1, height=1):
    [minx, miny, maxx, maxy] = hgut.tile_bounds(tileset_info, z, x, y, width, height)
    # print("x", x, 'y', y, width, height)
    max_per_tile = 30

    # get the entire region
    df = df.query(
        "{} < {} & {} < {} & {} < {} & {} < {}".format(
            minx, x_column, x_column, maxx, miny, y_column, y_column, maxy
        )
    )

    tiles = []
    if width > 1 or height > 1:
        for i in range(width):
            for j in range(height):
                # print("i:", i, "j:", j)
                [minx, miny, maxx, maxy] = hgut.tile_bounds(
                    tileset_info, z, x + i, y + j
                )

                data = df.query(
                    "{} < {} & {} < {} & {} < {} & {} < {}".format(
                        minx, x_column, x_column, maxx, miny, y_column, y_column, maxy
                    )
                )
                tiles += [((z, x + i, y + j), data[:max_per_tile])]
    else:
        return [((z, x, y), df[:max_per_tile])]

    return tiles



import clodius.array as ct
import math
import numpy as np


def get_tileset_info(hdf_file):
    """
    Get information about the tileset.

    :param hdf_file: A file handle for an HDF5 file (h5py.File('...'))
    """
    d = hdf_file["meta"]

    if "min-pos" in d.attrs:
        min_pos = d.attrs["min-pos"]
    else:
        min_pos = 0

    if "max-pos" in d.attrs:
        max_pos = d.attrs["max-pos"]
    else:
        max_pos = d.attrs["max-length"]

    return {
        "max_pos": max_pos,
        "min_pos": min_pos,
        "max_width": d.attrs["max-width"],
        "max_zoom": d.attrs["max-zoom"],
        "tile_size": d.attrs["tile-size"],
    }


def bisect_left(a, x, lo=0, hi=None, comparator=None):
    """Bisect_left with with an additional comparator.

    Based on the bisect_left function from the python bisect module.

    Return the index where to insert item x in list a, assuming a is sorted.

    The return value i is such that all e in a[:i] have e < x, and all e in
    a[i:] have e >= x.  So if x already appears in the list, a.insert(x) will
    insert just before the leftmost x already there.

    Optional args lo (default 0) and hi (default len(a)) bound the
    slice of a to be searched.

    Args:
        a (array): The array to bisect
        x (object): The object to find the insertion point of
        lo (int): The starting index of items to search in a
        hi (int): The end index of items to search in a
        comparator (function(a,b)): A way to compare objects
    """

    if lo < 0:
        raise ValueError("lo must be non-negative")
    if hi is None:
        hi = len(a)
    while lo < hi:
        mid = (lo + hi) // 2
        if comparator(a[mid], x) < 0:
            lo = mid + 1
        else:
            hi = mid
    return lo


def bisect_right(a, x, lo=0, hi=None, comparator=None):
    """Bisect_right with with an additional comparator.

    Based on the bisect_right function from the python bisect module.

    Return the index where to insert item x in list a, assuming a is sorted.

    The return value i is such that all e in a[:i] have e <= x, and all e in
    a[i:] have e > x.  So if x already appears in the list, a.insert(x) will
    insert just after the rightmost x already there.

    Optional args lo (default 0) and hi (default len(a)) bound the
    slice of a to be searched.


    Args:
        a (array): The array to bisect
        x (object): The object to find the insertion point of
        lo (int): The starting index of items to search in a
        hi (int): The end index of items to search in a
        comparator (function(a,b)): A way to compare objects
    """
    if lo < 0:
        raise ValueError("lo must be non-negative")
    if hi is None:
        hi = len(a)
    while lo < hi:
        mid = (lo + hi) // 2
        if comparator(x, a[mid]) < 0:
            hi = mid
        else:
            lo = mid + 1
    return lo


def get_discrete_data(hdf_file, z, x):
    """
    Get a discrete set of data from an hdf_tile file.

    Args:
        hdf_file (h5py.File): File handle for the file containing the information
        z (int): The zoom level of this tile
        x (int): The x position of this tile

    Returns:
        A 2D array of entries at that position. It is assumed that names of the
        columns are known.
    """
    # is the title within the range of possible tiles
    if x > 2 ** z:
        print("OUT OF RIGHT RANGE")
        return []
    if x < 0:
        print("OUT OF LEFT RANGE")
        return []

    d = hdf_file["meta"]
    tile_size = int(d.attrs["tile-size"])
    # max_length = int(d.attrs['max-length'])
    max_zoom = int(d.attrs["max-zoom"])
    # max_width = tile_size * 2 ** max_zoom

    # f is an array of data (e.g. [['34','53', 'x'],['48','57', 'y']] )
    # where the first two columns indicate the start and end points
    f = hdf_file[str(z)]

    # the tile width depends on the zoom level, lower zoom tiles encompass
    # a broader swatch of the data
    tile_width = tile_size * 2 ** (max_zoom - z)
    tile_start = x * tile_width
    tile_end = tile_start + tile_width

    # We need a way to compare data points that aren't numbers
    # (e.g. the arrays that are in f)
    def comparator_start(a, b):
        return int(a[0]) - int(b[0])

    def comparator_end(a, b):
        return int(a[1]) - int(b[1])

    tile_data_start = bisect_left(f, [tile_start], comparator=comparator_start)
    tile_data_end = bisect_right(f, [tile_end], comparator=comparator_start)

    return f[tile_data_start:tile_data_end]


def get_data(hdf_file, z, x):
    """
    Return a tile from an hdf_file.

    :param hdf_file: A file handle for an HDF5 file (h5py.File('...'))
    :param z: The zoom level
    :param x: The x position of the tile
    """

    # is the title within the range of possible tiles
    if x > 2 ** z:
        print("OUT OF RIGHT RANGE")
        return []
    if x < 0:
        print("OUT OF LEFT RANGE")
        return []

    d = hdf_file["meta"]

    tile_size = int(d.attrs["tile-size"])
    zoom_step = int(d.attrs["zoom-step"])
    max_zoom = int(d.attrs["max-zoom"])
    max_width = tile_size * 2 ** max_zoom

    if "max-position" in d.attrs:
        max_position = int(d.attrs["max-position"])
    else:
        max_position = max_width

    rz = max_zoom - z
    # tile_width = max_width / 2**z

    # because we only store some a subsection of the zoom levels
    next_stored_zoom = zoom_step * math.floor(rz / zoom_step)
    zoom_offset = rz - next_stored_zoom

    # the number of entries to aggregate for each new value
    num_to_agg = 2 ** zoom_offset
    total_in_length = tile_size * num_to_agg

    # which positions we need to retrieve in order to dynamically aggregate
    start_pos = int((x * 2 ** zoom_offset * tile_size))
    end_pos = int(start_pos + total_in_length)

    # print("max_position:", max_position)
    max_position = int(max_position / 2 ** next_stored_zoom)
    # print("new max_position:", max_position)

    """
    print("start_pos:", start_pos)
    print("end_pos:", end_pos)
    print("next_stored_zoom", next_stored_zoom)
    print("max_position:", int(max_position))
    """

    f = hdf_file["values_" + str(int(next_stored_zoom))]

    if start_pos > max_position:
        # we want a tile that's after the last bit of data
        a = np.zeros(end_pos - start_pos)
        a.fill(np.nan)
        ret_array = ct.aggregate(a, int(num_to_agg))
    elif start_pos < max_position and max_position < end_pos:
        a = f[start_pos:end_pos][:]
        a[max_position + 1 : end_pos] = np.nan
        ret_array = ct.aggregate(a, int(num_to_agg))
    else:
        ret_array = ct.aggregate(f[start_pos:end_pos], int(num_to_agg))

    """
    print("ret_array:", f[start_pos:end_pos])
    print('ret_array:', ret_array)
    """
    # print('nansum', np.nansum(ret_array))

    # check to see if we counted the number of NaN values in the given
    # interval

    f_nan = None
    if "nan_values_" + str(int(next_stored_zoom)) in hdf_file:
        f_nan = hdf_file["nan_values_" + str(int(next_stored_zoom))]
        nan_array = ct.aggregate(f_nan[start_pos:end_pos], int(num_to_agg))
        num_aggregated = 2 ** (max_zoom - z)

        num_vals_array = np.zeros(len(nan_array))
        num_vals_array.fill(num_aggregated)
        num_summed_array = num_vals_array - nan_array

        averages_array = ret_array / num_summed_array

        return averages_array

    return ret_array



import click
from . import cli
import clodius.chromosomes as cch
import clodius.multivec as cmv
import h5py
import math
import negspy.coordinates as nc
import numpy as np
import os
import os.path as op
import scipy.misc as sm
import tempfile

import ast


def epilogos_bedline_to_vector(bedlines, row_infos=None):
    """
    Convert a line from an epilogos bedfile to vector format.

    Parameters
    -----------
    bedline: [string,....]
        A line from a bedfile broken up into its constituent parts
        (e.g. ["chr1", "1000", "2000", "[1,2,34,5]"])

    Returns
    -------
    An array containing the values associated with that line
    """
    bedline = bedlines[0]
    parts = bedline.strip().split("\t")
    # extract the state values e.g. [...,[0,14],[0.56,15]]
    array_str = parts[3].split(":")[-1]

    # make sure they're ordered by their index
    array_val = sorted(ast.literal_eval(array_str), key=lambda x: x[1])
    states = [v[0] for v in array_val]

    chrom = parts[0]
    start = int(parts[1])
    end = int(parts[2])

    return (chrom, start, end, states)


def states_bedline_to_vector(bedlines, states_dic):
    """
    Convert a line from a bedfile containing states in categorical data to vector format.

    Parameters
    ----------

    bedline: [string,...]
        A line form a bedfile broken up into its contituent parts
        (e.g. ["chr1", "1000", "2000", "state"]))


    states_dic: {'key':val,...}
        A dictionary containing the states in the file with a corresponding value
        (e.g. {'state1_name': 1, 'state2_name': 2,...})

    Returns
    -------

    Four variables containing the values associated with that line: chrom, start, end, states_vector
    (e.g. chrom = "chr1", start = 1000, end = 2000, states_vector = [1,0,0,0])
    """
    # we support passing in multiple bed files for multivec creation from
    # other file types, but this one only supports a single file so just
    # assume that a single file is passed in
    bedline = bedlines[0]
    parts = bedline.strip().split("\t")
    chrom = parts[0]
    start = int(parts[1])
    end = int(parts[2])
    state = states_dic[parts[3]]

    states_vector = [1 if index == state else 0 for index in range(len(states_dic))]

    return (chrom, start, end, states_vector)


@cli.group()
def convert():
    """
    Aggregate a data file so that it stores the data at multiple
    resolutions.
    """
    pass


def _bedgraph_to_multivec(
    filepaths,
    output_file,
    assembly,
    chrom_col,
    from_pos_col,
    to_pos_col,
    value_col,
    has_header,
    chunk_size,
    nan_value,
    chromsizes_filename,
    starting_resolution,
    num_rows,
    format,
    row_infos_filename,
    tile_size,
    method,
):
    print("chrom_col:", chrom_col)

    with tempfile.TemporaryDirectory() as td:
        print("temporary dir:", td)

        temp_file = op.join(td, "temp.mv5")
        f_out = h5py.File(temp_file, "w")

        (chrom_info, chrom_names, chrom_sizes) = cch.load_chromsizes(
            chromsizes_filename, assembly
        )

        if row_infos_filename is not None:
            with open(row_infos_filename, "r") as fr:
                row_infos = [l.strip().encode("utf8") for l in fr]

        else:
            row_infos = None

        for chrom in chrom_info.chrom_order:
            f_out.create_dataset(
                chrom,
                (
                    math.ceil(chrom_info.chrom_lengths[chrom] / starting_resolution),
                    num_rows * len(filepaths),
                ),
                fillvalue=np.nan,
                compression="gzip",
            )

        def bedline_to_chrom_start_end_vector(bedlines, row_infos=None):
            chrom_set = set()
            start_set = set()
            end_set = set()
            all_vector = []

            for bedline in bedlines:
                parts = bedline.strip().split()
                chrom = parts[chrom_col - 1]
                start = int(parts[from_pos_col - 1])
                end = int(parts[to_pos_col - 1])
                vector = [
                    float(f) if not f == "NA" else np.nan
                    for f in parts[value_col - 1 : value_col - 1 + num_rows]
                ]
                chrom_set.add(chrom)
                start_set.add(start)
                end_set.add(end)

                if len(chrom_set) > 1:
                    raise ValueError(
                        "Chromosomes don't match in these lines:", bedlines
                    )
                if len(start_set) > 1:
                    raise ValueError(
                        "Start positions don't match in these lines:", bedlines
                    )
                if len(end_set) > 1:
                    raise ValueError(
                        "End positions don't match in these lines:", bedlines
                    )
                all_vector += vector

            return (
                list(chrom_set)[0],
                list(start_set)[0],
                list(end_set)[0],
                all_vector,
            )

        if format == "epilogos":
            cmv.bedfile_to_multivec(
                filepaths,
                f_out,
                epilogos_bedline_to_vector,
                starting_resolution,
                has_header,
                chunk_size,
                num_rows,
            )
        elif format == "states":
            assert (
                row_infos is not None
            ), "A row_infos file must be provided for --format = 'states' "
            states_names = [lne.decode("utf8").split("\t")[0] for lne in row_infos]
            states_dic = {states_names[x]: x for x in range(len(row_infos))}

            cmv.bedfile_to_multivec(
                filepaths,
                f_out,
                states_bedline_to_vector,
                starting_resolution,
                has_header,
                chunk_size,
                num_rows,
                states_dic,
            )
        else:
            cmv.bedfile_to_multivec(
                filepaths,
                f_out,
                bedline_to_chrom_start_end_vector,
                starting_resolution,
                has_header,
                chunk_size,
                num_rows,
            )

        f_out.close()
        tf = temp_file
        f_in = h5py.File(tf, "r")

        if output_file is None:
            output_file = op.splitext(filepaths[0])[0] + ".multires.mv5"
        print("output_file:", output_file)

        # Override the output file if it existts
        if op.exists(output_file):
            os.remove(output_file)

        if method == "logsumexp":

            def agg(x):
                # newshape = (x.shape[2], -1, 2)
                # b = x.T.reshape((-1,))

                a = x.T.reshape((x.shape[1], -1, 2))

                # this is going to be an odd way to get rid of nan
                # values
                orig_shape = a.shape
                na = a.reshape((-1,))

                SMALL_NUM = -1e8
                NAN_THRESHOLD_NUM = SMALL_NUM / 100

                if np.nanmin(na) < NAN_THRESHOLD_NUM:
                    raise ValueError(
                        "Error removing nan's when running logsumexp aggregation"
                    )

                na[np.isnan(na)] = SMALL_NUM
                na = na.reshape(orig_shape)
                res = sm.logsumexp(a, axis=2).T

                nres = res.reshape((-1,))
                # print("nres:", np.nansum(nres < NAN_THRESHOLD_NUM))
                nres[nres < NAN_THRESHOLD_NUM] = np.nan
                res = nres.reshape(res.shape)

                # print("res:", np.nansum(res.reshape((-1,))))

                return res

        else:

            def agg(x):
                return x.T.reshape((x.shape[1], -1, 2)).sum(axis=2).T

        if format == "states":
            states_row_infos = [
                state_name.encode("utf8") for state_name in states_names
            ]
            cmv.create_multivec_multires(
                f_in,
                chromsizes=zip(chrom_names, chrom_sizes),
                agg=agg,
                starting_resolution=starting_resolution,
                tile_size=tile_size,
                output_file=output_file,
                row_infos=states_row_infos,
            )
        else:
            cmv.create_multivec_multires(
                f_in,
                chromsizes=zip(chrom_names, chrom_sizes),
                agg=agg,
                starting_resolution=starting_resolution,
                tile_size=tile_size,
                output_file=output_file,
                row_infos=row_infos,
            )


@convert.command()
@click.argument("filepaths", metavar="FILEPATHS", nargs=-1)
@click.option(
    "--output-file",
    "-o",
    default=None,
    help="The default output file name to use. If this isn't"
    "specified, clodius will replace the current extension"
    "with .hitile",
)
@click.option(
    "--assembly",
    "-a",
    help="The genome assembly that this file was created against",
    type=click.Choice(nc.available_chromsizes()),
    default="hg19",
)
@click.option(
    "--chromosome-col",
    help="The column number (1-based) which contains the chromosome " "name",
    default=1,
    type=int,
)
@click.option(
    "--from-pos-col",
    help="The column number (1-based) which contains the starting " "position",
    default=2,
    type=int,
)
@click.option(
    "--to-pos-col",
    help="The column number (1-based) which contains the ending" "position",
    default=3,
    type=int,
)
@click.option(
    "--value-col",
    help="The column number (1-based) which contains the actual value" "position",
    default=4,
    type=int,
)
@click.option(
    "--has-header/--no-header",
    help="Does this file have a header that we should ignore",
    default=False,
)
@click.option(
    "--chunk-size", help="The size of the chunks to read in at once", default=1e5
)
@click.option(
    "--nan-value", help="The string to use as a NaN value", type=str, default=None
)
@click.option(
    "--chromsizes-filename",
    help="A file containing chromosome sizes and order",
    default=None,
)
@click.option(
    "--starting-resolution",
    help="The base resolution of the data. Used to determine how much space to allocate"
    " in the multivec file",
    default=1,
)
@click.option(
    "--num-rows",
    help="The number of rows at each position in the multivec format",
    default=1,
)
@click.option(
    "--format",
    type=click.Choice(["default", "epilogos", "states"]),
    help="'default':chr start end state1_value state2_value, etc;"
    "'epilogos': chr start end [[state1_value, state1_num],[state2_value, state2_num],[etc]];"
    "'states': chr start end state_name",
    default="default",
)
@click.option(
    "--row-infos-filename",
    help="A file containing the names of the rows in the multivec file",
    default=None,
)
@click.option(
    "--tile-size",
    "-t",
    default=256,
    help="The number of data points in each tile."
    "Used to determine the number of zoom levels"
    "to create.",
)
@click.option(
    "--method",
    help="The method used to aggregate values (e.g. sum, average...)",
    type=click.Choice(["sum", "logsumexp"]),
    default="sum",
)
def bedfile_to_multivec(
    filepaths,
    output_file,
    assembly,
    chromosome_col,
    from_pos_col,
    to_pos_col,
    value_col,
    has_header,
    chunk_size,
    nan_value,
    chromsizes_filename,
    starting_resolution,
    num_rows,
    format,
    row_infos_filename,
    tile_size,
    method,
):
    _bedgraph_to_multivec(
        filepaths,
        output_file,
        assembly,
        chromosome_col,
        from_pos_col,
        to_pos_col,
        value_col,
        has_header,
        chunk_size,
        nan_value,
        chromsizes_filename,
        starting_resolution,
        num_rows,
        format,
        row_infos_filename,
        tile_size,
        method,
    )



import clodius.tiles.nplabels as ctn
import h5py


def test_tiles():
    filename = "data/labels.h5"
    f = h5py.File(filename, "r")

    array = f["labels"]

    tiles = ctn.tiles(array, 0, 0, None)

    assert tiles[0]["x"] == 0
    assert tiles[-1]["x"] == 86992

    tiles = ctn.tiles(array, 2, 1, None)

    assert tiles[0]["x"] == 32768
    assert tiles[-1]["x"] == 65535



import functools as ft
import clodius.tiles.density as hgde
import clodius.tiles.utils as hgut


def test_bundled_tiles():
    filename = "data/points_density.h5"

    tiles = hgut.bundled_tiles_wrapper_2d(
        ["c.2.0.0", "c.2.1.1", "c.2.0.1", "c.2.1.0"], ft.partial(hgde.tiles, filename)
    )
    assert len(tiles) == 4



import base64
import h5py
import math
import numpy as np
import os
import os.path as op


def array_to_hitile(
    old_data, filename, zoom_step=8, chunks=(1e6,), agg_function=np.sum
):
    """
    Downsample a dataset so that it's compatible with HiGlass (filetype: hitile, datatype: vector)

    Parameters
    ----------
    old_data: np.array
        A numpy array containing the data to be downsampled
    filename: string
        The output filename where the resulting multi-resolution
        data will be stored.
    zoom_step: int
        The number of zoom levels to skip when aggregating
    """
    import dask.array as da

    if op.exists(filename):
        os.remove(filename)

    f_new = h5py.File(filename, "w")

    tile_size = 1024
    max_pos = len(old_data)

    # we store every n'th zoom level
    zoom_factor = 2 ** zoom_step

    max_zoom = math.ceil(math.log(max_pos / tile_size) / math.log(2))

    meta = f_new.create_dataset("meta", (1,), dtype="f")
    meta.attrs["tile-size"] = tile_size
    meta.attrs["zoom-step"] = zoom_step
    meta.attrs["max-length"] = max_pos
    meta.attrs["max-zoom"] = max_zoom

    meta.attrs["max-width"] = tile_size * 2 ** max_zoom

    min_data = da.from_array(old_data, chunks)
    max_data = da.from_array(old_data, chunks)
    old_data = da.from_array(old_data, chunks)

    for z in range(0, max_zoom, zoom_step):
        values_dset = f_new.require_dataset(
            "values_" + str(z), (len(old_data),), dtype="f", compression="gzip"
        )

        mins_dset = f_new.require_dataset(
            "mins_" + str(z), (len(old_data),), dtype="f", compression="gzip"
        )
        maxs_dset = f_new.require_dataset(
            "maxs_" + str(z), (len(old_data),), dtype="f", compression="gzip"
        )

        da.store(old_data, values_dset)
        da.store(min_data, mins_dset)
        da.store(max_data, maxs_dset)
        # f_new['values_' + str(z)][:] = old_data

        # see if we need to pad the end of the dataset
        # if so, use the previous last value
        if len(old_data) % zoom_factor != 0:
            old_data = da.concatenate(
                (old_data, [old_data[-1]] * (zoom_factor - len(old_data) % zoom_factor))
            )
            min_data = da.concatenate(
                (min_data, [max_data[-1]] * (zoom_factor - len(min_data) % zoom_factor))
            )
            max_data = da.concatenate(
                (max_data, [max_data[-1]] * (zoom_factor - len(max_data) % zoom_factor))
            )

        # aggregate the data by summing adjacent datapoints
        # sys.stdout.write('summing...')
        # sys.stdout.flush()
        # print("fdsdsfs:", math.ceil(len(old_data) / zoom_factor), zoom_factor)
        # print("chunks:", chunks, zoom_factor, 'len:', len(old_data))

        old_data = old_data.rechunk(chunks)
        min_data = old_data.rechunk(chunks)
        max_data = old_data.rechunk(chunks)
        # print('zoom_factor', zoom_factor, old_data.shape)

        old_data = da.coarsen(agg_function, old_data, {0: zoom_factor})
        min_data = da.coarsen(np.min, max_data, {0: zoom_factor})
        max_data = da.coarsen(np.max, max_data, {0: zoom_factor})

        # reshape( (math.ceil(len(old_data) / zoom_factor), zoom_factor)).sum(axis=1)
        # sys.stdout.write(' done\n')
        # sys.stdout.flush()

        """
        if len(old_data) < 10000:
            plt.plot(old_data)
        """

    # plt.plot(old_data)
    f_new.close()


def aggregate(a, num_to_agg):
    if len(a) % num_to_agg != 0:
        a = np.concatenate((a, [a[-1]] * (num_to_agg - len(a) % num_to_agg)))

    return a.reshape((math.ceil(len(a) / num_to_agg), num_to_agg)).sum(axis=1)


def aggregate_min(a, num_to_agg):
    if len(a) % num_to_agg != 0:
        a = np.concatenate((a, [a[-1]] * (num_to_agg - len(a) % num_to_agg)))

    return a.reshape((math.ceil(len(a) / num_to_agg), num_to_agg)).min(axis=1)


def aggregate_max(a, num_to_agg):
    if len(a) % num_to_agg != 0:
        a = np.concatenate((a, [a[-1]] * (num_to_agg - len(a) % num_to_agg)))

    return a.reshape((math.ceil(len(a) / num_to_agg), num_to_agg)).max(axis=1)


def get_data(hdf_file, z, x):
    """
    Return a tile from an hdf_file.

    :param hdf_file: A file handle for an HDF5 file (h5py.File('...'))
    :param z: The zoom level
    :param x: The x position of the tile
    """

    # is the title within the range of possible tiles
    if x > 2 ** z:
        print("OUT OF RIGHT RANGE")
        return ([], [], [])
    if x < 0:
        print("OUT OF LEFT RANGE")
        return ([], [], [])

    d = hdf_file["meta"]

    tile_size = int(d.attrs["tile-size"])
    zoom_step = int(d.attrs["zoom-step"])
    max_zoom = int(d.attrs["max-zoom"])

    max_width = tile_size * 2 ** max_zoom

    if "max-position" in d.attrs:
        max_position = int(d.attrs["max-position"])
    else:
        max_position = max_width

    rz = max_zoom - z
    # tile_width = max_width / 2**z

    # because we only store some a subsection of the zoom levels
    next_stored_zoom = zoom_step * math.floor(rz / zoom_step)
    zoom_offset = rz - next_stored_zoom

    # the number of entries to aggregate for each new value
    num_to_agg = 2 ** zoom_offset
    total_in_length = tile_size * num_to_agg

    # which positions we need to retrieve in order to dynamically aggregate
    start_pos = int((x * 2 ** zoom_offset * tile_size))
    end_pos = int(start_pos + total_in_length)

    # print("max_position:", max_position)
    max_position = int(max_position / 2 ** next_stored_zoom)
    # print("new max_position:", max_position)

    # print("start_pos:", start_pos)
    # print("end_pos:", end_pos)
    # print("next_stored_zoom", next_stored_zoom)
    # print("max_position:", int(max_position))

    f = hdf_file["values_" + str(int(next_stored_zoom))]
    f_min = hdf_file["mins_" + str(int(next_stored_zoom))]
    f_max = hdf_file["maxs_" + str(int(next_stored_zoom))]

    if start_pos > max_position:
        # we want a tile that's after the last bit of data
        a = np.zeros(end_pos - start_pos)
        a.fill(np.nan)

        a_min = np.zeros(end_pos - start_pos)
        a_min.fill(np.nan)

        # umm, I don't think this needs to be here since
        # everything should be nan
        ret_array = aggregate(a, int(num_to_agg))
        min_array = aggregate_min(a_min, int(num_to_agg))
        # In the line below, "a_max" is undefined, so this would not work:
        # max_array = aggregate_max(a_max, int(num_to_agg))
    elif start_pos < max_position and max_position < end_pos:
        a = f[start_pos:end_pos][:]
        a[max_position + 1 : end_pos] = np.nan

        a_min = f_min[start_pos:end_pos][:]
        a_min[max_position + 1 : end_pos] = np.nan

        a_max = f_max[start_pos:end_pos][:]
        a_max[max_position + 1 : end_pos] = np.nan

        ret_array = aggregate(a, int(num_to_agg))
        min_array = aggregate_min(a_min, int(num_to_agg))
        max_array = aggregate_max(a_max, int(num_to_agg))
    else:
        ret_array = aggregate(f[start_pos:end_pos], int(num_to_agg))
        min_array = aggregate_min(f_min[start_pos:end_pos], int(num_to_agg))
        max_array = aggregate_max(f_max[start_pos:end_pos], int(num_to_agg))

    # print("ret_array:", f[start_pos:end_pos])
    # print('ret_array:', ret_array)

    # print('nansum', np.nansum(ret_array))

    # check to see if we counted the number of NaN values in the given
    # interval

    f_nan = None
    if "nan_values_" + str(int(next_stored_zoom)) in hdf_file:
        f_nan = hdf_file["nan_values_" + str(int(next_stored_zoom))]
        nan_array = aggregate(f_nan[start_pos:end_pos], int(num_to_agg))
        num_aggregated = 2 ** (max_zoom - z)

        num_vals_array = np.zeros(len(nan_array))
        num_vals_array.fill(num_aggregated)
        num_summed_array = num_vals_array - nan_array

        averages_array = ret_array / num_summed_array

        return (averages_array, min_array, max_array)

    return (ret_array, min_array, max_array)


def tileset_info(hitile_path):
    """
    Get the tileset info for a hitile file.

    Parameters
    ----------
    hitile_path: string
        The path to the hitile file

    Returns
    -------
    tileset_info: {'min_pos': [],
                    'max_pos': [],
                    'tile_size': 1024,
                    'max_zoom': 7
                    }
    """
    hdf_file = h5py.File(hitile_path, "r")

    d = hdf_file["meta"]

    if "min-pos" in d.attrs:
        min_pos = d.attrs["min-pos"]
    else:
        min_pos = 0

    if "max-pos" in d.attrs:
        max_pos = d.attrs["max-pos"]
    else:
        max_pos = d.attrs["max-length"]

    return {
        "max_pos": [int(max_pos)],
        "min_pos": [int(min_pos)],
        "max_width": 2 ** math.ceil(math.log(max_pos - min_pos) / math.log(2)),
        "max_zoom": int(d.attrs["max-zoom"]),
        "tile_size": int(d.attrs["tile-size"]),
    }


def tiles(filepath, tile_ids):
    """
    Generate tiles from a hitile file.

    Parameters
    ----------
    tileset: tilesets.models.Tileset object
        The tileset that the tile ids should be retrieved from
    tile_ids: [str,...]
        A list of tile_ids (e.g. xyx.0.0) identifying the tiles
        to be retrieved

    Returns
    -------
    tile_list: [(tile_id, tile_data),...]
        A list of tile_id, tile_data tuples
    """
    generated_tiles = []

    for tile_id in tile_ids:
        tile_id_parts = tile_id.split(".")
        tile_position = list(map(int, tile_id_parts[1:3]))

        (dense, mins, maxs) = get_data(
            h5py.File(filepath), tile_position[0], tile_position[1]
        )

        """
        if len(dense):
            max_dense = max(dense)
            min_dense = min(dense)
        else:
            max_dense = 0
            min_dense = 0
        has_nan = len([d for d in dense if np.isnan(d)]) > 0
        if (
            not has_nan and
            max_dense > min_f16 and max_dense < max_f16 and
            min_dense > min_f16 and min_dense < max_f16
        ):
            tile_value = {
                'dense': base64.b64encode(dense.astype('float16')).decode('utf-8'),
                'mins': base64.b64encode(mins.astype('float16')).decode('utf-8'),
                'maxs': base64.b64encode(mins.astype('float16')).decode('utf-8'),
                'dtype': 'float16'
            }
        else:
        """
        tile_value = {
            "dense": base64.b64encode(dense.astype("float32")).decode("utf-8"),
            "mins": base64.b64encode(mins.astype("float32")).decode("utf-8"),
            "maxs": base64.b64encode(maxs.astype("float32")).decode("utf-8"),
            "dtype": "float32",
        }

        generated_tiles += [(tile_id, tile_value)]

    return generated_tiles



import math
import numpy as np
import pysam


def abs2genomic(chromsizes, start_pos, end_pos):
    abs_chrom_offsets = np.r_[0, np.cumsum(chromsizes)]
    cid_lo, cid_hi = (
        np.searchsorted(abs_chrom_offsets, [start_pos, end_pos], side="right") - 1
    )
    rel_pos_lo = start_pos - abs_chrom_offsets[cid_lo]
    rel_pos_hi = end_pos - abs_chrom_offsets[cid_hi]
    start = rel_pos_lo
    for cid in range(cid_lo, cid_hi):
        yield cid, start, chromsizes[cid]
        start = 0
    yield cid_hi, start, rel_pos_hi


def load_reads(samfile, start_pos, end_pos, chrom_order=None):
    """
    Sample reads from the specified region, assuming that the chromosomes
    are ordered in some fashion. Returns an list of pysam reads

    Parameters:
    -----------
    samfile: pysam.AlignmentFile
        A pysam entry into an indexed bam file
    start_pos: int
        The start position of the sampled region
    end_pos: int
        The end position of the sampled region
    chrom_order: ['chr1', 'chr2',...]
        A listing of chromosome names to use as the order

    Returns
    -------
    reads: [read1, read2...]
        The list of in the sampled regions
    """
    # if chromorder is not None...
    # specify the chromosome order for the fetched reads

    references = np.array(samfile.references)
    lengths = np.array(samfile.lengths)
    abs_chrom_offsets = np.r_[0, np.cumsum(lengths)]

    if chrom_order:
        chrom_order = np.array(chrom_order)
        chrom_order_ixs = np.nonzero(np.in1d(references, chrom_order))
        lengths = lengths[chrom_order_ixs]

    results = {
        "id": [],
        "from": [],
        "to": [],
        "md": [],
        "chrName": [],
        "chrOffset": [],
        "cigar": [],
    }

    for cid, start, end in abs2genomic(lengths, start_pos, end_pos):
        chr_offset = int(abs_chrom_offsets[cid])

        seq_name = f"{references[cid]}"
        reads = samfile.fetch(seq_name, start, end)

        for read in reads:
            if read.is_unmapped:
                continue
            # query_seq = read.query_sequence

            # differences = []

            # try:
            #     for counter, (qpos, rpos, ref_base) in enumerate(read.get_aligned_pairs(with_seq=True)):
            #         # inferred from the pysam source code:
            #         # https://github.com/pysam-developers/pysam/blob/3defba98911d99abf8c14a483e979431f069a9d2/pysam/libcalignedsegment.pyx
            #         # and GitHub issue:
            #         # https://github.com/pysam-developers/pysam/issues/163
            #         #print('qpos, rpos, ref_base', qpos, rpos, ref_base)
            #         if rpos is None:
            #             differences += [(qpos, 'I')]
            #         elif qpos is None:
            #             differences += [(counter, 'D')]
            #         elif ref_base.islower():
            #             differences += [(qpos, query_seq[qpos], ref_base)]
            # except ValueError as ve:
            #     # probably lacked an MD string
            #     pass
            try:
                results["id"] += [read.query_name]
                results["from"] += [int(read.reference_start + chr_offset)]
                results["to"] += [int(read.reference_end + chr_offset)]
                results["chrName"] += [read.reference_name]
                results["chrOffset"] += [chr_offset]
                results["cigar"] += [read.cigarstring]
            except:
                raise

            try:
                results["md"] += [read.get_tag("MD")]
            except KeyError:
                results["md"] += [""]
                continue

    return results


def tileset_info(filename):
    """
    Get the tileset info for a bam file

    Parameters
    ----------
    tileset: tilesets.models.Tileset object
        The tileset that the tile ids should be retrieved from

    Returns
    -------
    tileset_info: {'min_pos': [],
                    'max_pos': [],
                    'tile_size': 1024,
                    'max_zoom': 7
                    }
    """
    samfile = pysam.AlignmentFile(filename)
    total_length = sum(samfile.lengths)

    tile_size = 256
    max_zoom = math.ceil(math.log(total_length / tile_size) / math.log(2))

    tileset_info = {
        "min_pos": [0],
        "max_pos": [total_length],
        "max_width": tile_size * 2 ** max_zoom,
        "tile_size": tile_size,
        "max_zoom": max_zoom,
    }

    return tileset_info


def tiles(filename, tile_ids, index_filename=None, max_tile_width=None):
    """
    Generate tiles from a bigwig file.

    Parameters
    ----------
    tileset: tilesets.models.Tileset object
        The tileset that the tile ids should be retrieved from
    tile_ids: [str,...]
        A list of tile_ids (e.g. xyx.0.0) identifying the tiles
        to be retrieved
    index_filename: str
        The name of the file containing the index
    max_tile_width: int
        How wide can each tile be before we return no data. This
        can be used to limit the amount of data returned.

    Returns
    -------
    tile_list: [(tile_id, tile_data),...]
        A list of tile_id, tile_data tuples
    """
    generated_tiles = []
    tsinfo = tileset_info(filename)
    samfile = pysam.AlignmentFile(filename, index_filename=index_filename)

    for tile_id in tile_ids:
        tile_id_parts = tile_id.split("|")[0].split(".")
        tile_position = list(map(int, tile_id_parts[1:3]))

        tile_width = tsinfo["max_width"] / 2 ** int(tile_position[0])

        if max_tile_width and tile_width >= max_tile_width:
            # this tile is larger than the max allowed
            return [
                (
                    tile_id,
                    {
                        "error": f"Tile too large, no data returned. Max tile size: {max_tile_width}"
                    },
                )
            ]
        else:
            start_pos = int(tile_position[1]) * tile_width
            end_pos = start_pos + tile_width

            tile_value = load_reads(samfile, start_pos=start_pos, end_pos=end_pos)
            generated_tiles += [(tile_id, tile_value)]

    return generated_tiles



from __future__ import print_function

import click.testing as clt
import clodius.cli.convert as ccc
import clodius.tiles.multivec as ctv
import os.path as op
import tempfile
import h5py

testdir = op.realpath(op.dirname(__file__))


def test_bedfile_to_multivec():
    runner = clt.CliRunner()
    input_file = op.join(testdir, "sample_data", "sample.multival.bed")
    chromsizes_file = op.join(testdir, "sample_data", "sample.multival.chrom.sizes")

    with tempfile.TemporaryDirectory() as tmp_dir:
        out_file = op.join(tmp_dir, "out.multivec")

        _ = runner.invoke(
            ccc.bedfile_to_multivec,
            [
                input_file,
                "--output-file",
                out_file,
                "--assembly",
                "hg38",
                "--num-rows",
                3,
                "--chromsizes-filename",
                chromsizes_file,
                "--starting-resolution",
                "1000",
            ],
        )

        # import traceback
        # a, b, tb = result.exc_info

        # print("exc_info:", result.exc_info)
        # print("result:", result)
        # print("result.output", result.output)
        # print("result.error", traceback.print_tb(tb))
        # print("Exception:", a,b)

        tsinfo = ctv.tileset_info(out_file)
        # print("tsinfo:", tsinfo)

        assert "resolutions" in tsinfo
        assert tsinfo["max_pos"][0] == 18000
        tile = ctv.get_single_tile(out_file, (0, 0))

        # input_file:
        # chr1    0   1000    1.0 2.0 3.0
        # chr1    1000    2000
        # chr2    5000    6000    20.0    30.0    40.0
        #
        # # input chromsizes
        # chr1  10000
        # chr2    8000

        # first row, first chrom first value
        assert len(tile) == 3

        assert tile[0][0] == 1.0
        assert tile[0][15] == 20.0

        assert tile[1][0] == 2.0
        assert tile[2][0] == 3.0
        assert tile[2][15] == 40.0


def test_load_multivec_tiles():
    op.join(testdir, "sample_data", "sample.bed.multires.mv5")
    # TODO: Make assertions about result


def test_states_format_befile_to_multivec():
    runner = clt.CliRunner()
    input_file = op.join(testdir, "sample_data", "states_format_input_testfile.bed.gz")
    rows_info_file = op.join(testdir, "sample_data", "states_format_test_row_infos.txt")
    tempfile.NamedTemporaryFile(delete=False)
    # TODO: Make assertions about result
    # print("input_file", input_file)

    result = runner.invoke(
        ccc.bedfile_to_multivec,
        [
            input_file,
            "--format",
            "states",
            "--row-infos-filename",
            rows_info_file,
            "--assembly",
            "hg38",
            "--starting-resolution",
            "200",
            "--num-rows",
            "10",
        ],
    )

    # import traceback
    a, b, tb = result.exc_info
    """
    print("exc_info:", result.exc_info)
    print("result:", result)
    print("result.output", result.output)
    print("result.error", traceback.print_tb(tb))
    print("Exception:", a,b)
    """


def test_ignore_bedfile_headers():
    runner = clt.CliRunner()
    input_file = op.join(testdir, "sample_data", "3_header_100_testfile.bed.gz")
    rows_info_file = op.join(testdir, "sample_data", "3_header_100_row_infos.txt")
    tempfile.NamedTemporaryFile(delete=False)
    # TODO: Make assertions about result

    result = runner.invoke(
        ccc.bedfile_to_multivec,
        [
            input_file,
            "--format",
            "states",
            "--row-infos-filename",
            rows_info_file,
            "--assembly",
            "hg19",
            "--starting-resolution",
            "200",
            "--num-rows",
            "15",
        ],
    )

    # import traceback
    a, b, tb = result.exc_info


def test_retain_lines():
    runner = clt.CliRunner()
    input_file = op.join(testdir, "sample_data", "sample2.multival.bed")
    chromsizes_file = op.join(testdir, "sample_data", "sample2.multival.chrom.sizes")
    row_infos_file = op.join(
        testdir, "sample_data", "states_format_test_row_infos_v2.txt"
    )

    with tempfile.TemporaryDirectory() as tmp_dir:
        out_file = op.join(tmp_dir, "out.multires.mv5")

        result = runner.invoke(
            ccc.bedfile_to_multivec,
            [
                input_file,
                "--output-file",
                out_file,
                "--format",
                "states",
                "--chromsizes-filename",
                chromsizes_file,
                "--starting-resolution",
                "1000",
                "--row-infos-filename",
                row_infos_file,
                "--num-rows",
                "3",
            ],
        )

        # import traceback
        a, b, tb = result.exc_info

        # input_file:
        # chr1    0   1000  State1
        # chr1    1000    10111  State2
        # chr2    5000    8000  State3
        #
        # # input chromsizes
        # chr1  10111
        # chr2    8000
        #
        # # input row_infos file
        # State1
        # State2
        # State3

        f = h5py.File(out_file, "r")
        # The last bin of chromosome 1 should contain the State2 Vector [0,1,0]
        assert f["resolutions"]["1000"]["values"]["chr1"][10][0] == 0.0
        assert f["resolutions"]["1000"]["values"]["chr1"][10][1] == 1.0
        assert f["resolutions"]["1000"]["values"]["chr1"][10][2] == 0.0



import math
import base64

import h5py
import numpy as np


def abs2genomic(chromsizes, start_pos, end_pos):
    """
    Convert absolute genomic sizes to genomic

    Parameters:
    -----------
    chromsizes: [1000,...]
        An array of the lengths of the chromosomes
    start_pos: int
        The starting genomic position
    end_pos: int
        The ending genomic position
    """
    abs_chrom_offsets = np.r_[0, np.cumsum(chromsizes)]
    cid_lo, cid_hi = (
        np.searchsorted(abs_chrom_offsets, [start_pos, end_pos], side="right") - 1
    )
    rel_pos_lo = start_pos - abs_chrom_offsets[cid_lo]
    rel_pos_hi = end_pos - abs_chrom_offsets[cid_hi]
    start = rel_pos_lo
    for cid in range(cid_lo, cid_hi):
        yield cid, start, chromsizes[cid]
        start = 0
    yield cid_hi, start, rel_pos_hi


def tiles(filename, tile_ids):
    """
    Retrieve multiple multivec tiles from tids.
    ----------
    filename: string
        The multires file containing the multivec data
    tile_ids: [str,...]
        A list of tile_ids (e.g. xyx.0.0) identifying the tiles
        to be retrieved
    """
    f16 = np.finfo("float16")
    f16_min, f16_max = f16.min, f16.max
    generated_tiles = []
    for tile_id in tile_ids:
        tile_pos = [int(i) for i in tile_id.split(".")[1:3]]
        ma = get_single_tile(filename, tile_pos)
        has_nan = np.isnan(ma).any()
        ma_max = ma.max() if ma.size else 0
        ma_min = ma.min() if ma.size else 0
        use_f16 = not has_nan and (ma_min > f16_min and ma_max < f16_max)
        ma = ma.astype(np.float16 if use_f16 else np.float32)
        ma_base64 = base64.b64encode(ma.ravel()).decode("utf-8")
        tile_value = {
            "dense": ma_base64,
            "dtype": "float16" if use_f16 else "float32",
            "shape": ma.shape,
        }
        generated_tiles.append((tile_id, tile_value))

    return generated_tiles


def get_single_tile(filename, tile_pos):
    """
    Retrieve a single multivec tile from a multires file
    Parameters
    ----------
    filename: string
        The multires file containing the multivec data
    tile_pos: (z, x)
        The zoom level and position of this tile
    """
    # t1 = time.time()
    tsinfo = tileset_info(filename)
    f = h5py.File(filename, "r")

    # print('tileset_info', tileset_info)
    # t2 = time.time()
    # which resolution does this zoom level correspond to?
    resolution = tsinfo["resolutions"][tile_pos[0]]
    tile_size = tsinfo["tile_size"]

    # where in the data does the tile start and end
    tile_start = tile_pos[1] * tile_size * resolution
    tile_end = tile_start + tile_size * resolution

    chromsizes = list(zip(f["chroms"]["name"], f["chroms"]["length"]))

    # dense = f['resolutions'][str(resolution)][tile_start:tile_end]
    dense = get_tile(f, chromsizes, resolution, tile_start, tile_end, tsinfo["shape"])
    # print("dense.shape", dense.shape)

    if len(dense) < tsinfo["tile_size"]:
        # if there aren't enough rows to fill this tile, add some zeros
        dense = np.vstack(
            [dense, np.zeros((tsinfo["tile_size"] - len(dense), tsinfo["shape"][1]))]
        )

    f.close()

    # t3 = time.time()
    # print("single time time: {:.2f} (tileset info: {:.2f}, open time: {:.2f})".format(t3 - t1, t15 - t1, t2 - t15))

    return dense.T


def get_tile(f, chromsizes, resolution, start_pos, end_pos, shape):
    """
    Get the tile value given the start and end positions and
    chromosome positions.

    Drop bins at the ends of chromosomes if those bins aren't
    full.

    Parameters:
    -----------
    f: h5py.File
        An hdf5 file containing the data
    chromsizes: [('chr1', 1000), ....]
        An array listing the chromosome sizes
    resolution: int
        The size of each bin, except for the last bin in each
        chromosome.
    start_pos: int
        The start_position of the interval to return
    end_pos: int
        The end position of the interval to return

    Returns
    -------
    return_vals: [...]
        A subset of the original genome-wide values containing
        the values for the portion of the genome that is visible.
    """
    binsize = resolution
    # print('binsize:', binsize)
    # print('start_pos:', start_pos, 'end_pos:', end_pos)
    # print("length:", end_pos - start_pos)
    # print('shape:', shape)

    # t0 = time.time()
    arrays = []
    count = 0

    # keep track of how much data has been returned in bins
    current_binned_data_position = 0
    current_data_position = 0

    num_added = 0
    total_length = 0

    for cid, start, end in abs2genomic([c[1] for c in chromsizes], start_pos, end_pos):
        n_bins = int(np.ceil((end - start) / binsize))
        total_length += end - start
        # print('cid', cid, start, end, 'tl:', total_length)

        try:
            # t1 = time.time()

            chrom = chromsizes[cid][0]

            offset = current_binned_data_position - current_data_position
            current_data_position += end - start

            count += 1

            start_pos = math.floor((start + offset) / binsize)
            end_pos = math.ceil(end / binsize)

            if start_pos >= end_pos:
                continue

            # print("start:", start, "end", end)
            # print("sp", start_pos * binsize, end_pos * binsize)
            # print('current_data_position:', current_data_position)
            # print('current_binned_data_position:', current_binned_data_position)
            # print('binsize:', binsize, 'resolution:', resolution)

            """
            if start_pos == end_pos:
                if current_data_position - current_binned_data_position > 0:
                    # adding this data as a single bin even though it's not large
                    # enough to cover one bin
                    # print('catching up')
                    end_pos += 1
                else:
                    # print('data smaller than the bin size', start, end, binsize)
                    continue
            """

            # print("offset:", offset, "start_pos", start_pos, end_pos)
            x = f["resolutions"][str(resolution)]["values"][chrom][start_pos:end_pos]
            current_binned_data_position += binsize * (end_pos - start_pos)

            # print("x:", x.shape)

            # drop the very last bin if it is smaller than the binsize
            """
            if len(x) > 1 and end == clen and clen % binsize != 0:
                # print("dropping")
                x = x[:-1]
            """

            if len(x):
                num_added += len(x)
                # print('cid:', cid, end-start, total_length, 'num_added:', num_added, 'x:', sum(x))

            # t2 = time.time()
            # print("time to fetch {}: {}".format(chrom, t2 - t1))
        except IndexError:
            # beyond the range of the available chromosomes
            # probably means we've requested a range of absolute
            # coordinates that stretch beyond the end of the genome
            # print('zeroes')
            x = np.zeros((n_bins, shape[1]))

        arrays.append(x)

    # print("total_length:", total_length)
    # print('arrays:', len(np.concatenate(arrays)))
    # t3 = time.time()
    # print("total fetch time:", t3 - t0)

    return np.concatenate(arrays)[: shape[0]]


def tileset_info(filename):
    """
    Return some information about this tileset that will
    help render it in on the client.

    Parameters
    ----------
    filename: str
      The filename of the h5py file containing the tileset info.

    Returns
    -------
    tileset_info: {}
      A dictionary containing the information describing
      this dataset
    """
    # t1 = time.time()
    f = h5py.File(filename, "r")
    # t2 = time.time()
    # a sorted list of resolutions, lowest to highest
    # awkward to write because a the numbers representing resolution
    # are datapoints / pixel so lower resolution is actually a higher
    # number
    resolutions = sorted([int(r) for r in f["resolutions"].keys()])[::-1]

    # the "leftmost" datapoint position
    # an array because higlass can display multi-dimensional
    # data
    min_pos = [0]
    max_pos = [int(sum(f["chroms"]["length"][:]))]

    # the "rightmost" datapoint position
    # max_pos = [len(f['resolutions']['values'][str(resolutions[-1])])]
    tile_size = int(f["info"].attrs["tile-size"])
    first_chrom = f["chroms"]["name"][0]

    shape = list(f["resolutions"][str(resolutions[0])]["values"][first_chrom].shape)
    shape[0] = tile_size

    # t3 = time.time()
    # print("tileset info time:", t3 - t2)

    tileset_info = {
        "resolutions": resolutions,
        "min_pos": min_pos,
        "max_pos": max_pos,
        "tile_size": tile_size,
        "shape": shape,
    }

    if "row_infos" in f["resolutions"][str(resolutions[0])].attrs:
        row_infos = f["resolutions"][str(resolutions[0])].attrs["row_infos"]
        tileset_info["row_infos"] = [r.decode("utf8") for r in row_infos]

    f.close()

    return tileset_info



import sqlite3


def tileset_info(db_file):
    conn = sqlite3.connect(db_file)
    cursor = conn.cursor()

    row = cursor.execute("SELECT * from tileset_info").fetchone()

    colnames = next(zip(*cursor.description))

    if "version" not in colnames:
        version = 1
    else:
        version = int(row[colnames.index("version")])

    if "header" not in colnames:
        header = ""
    else:
        header = row[colnames.index("header")]

    ts_info = {
        "zoom_step": row[0],
        "max_length": row[1],
        "assembly": row[2],
        "chrom_names": row[3],
        "chrom_sizes": row[4],
        "tile_size": row[5],
        "max_zoom": row[6],
        "max_width": row[7],
        "min_pos": [1],
        "max_pos": [row[1]],
        "header": header,
        "version": version,
    }
    conn.close()

    return ts_info


def tiles(filepath, tile_ids):
    """
    Generate tiles from this dataset.

    Parameters
    ----------
    filepath: str
        The filename of the sqlite db file
    tile_ids: [str...]
        A list of tile ids of the form

    Returns
    -------
    tiles: [(tile_id, tile_value),...]
        A list of values indexed by the tile position
    """
    to_return = []

    for tile_id in tile_ids:
        # tile_option_parts = tile_id.split('|')[1:]
        tile_no_options = tile_id.split("|")[0]
        parts = tile_no_options.split(".")

        zoom = int(parts[1])
        xpos = int(parts[2])

        extra_zoom = 0
        new_rows = {}
        new_rows = []

        for j in range(2 ** extra_zoom):
            # the old rows are indexed by the higher
            # resolution tile numbers
            higher_xpos = 2 ** extra_zoom * xpos + j
            old_rows = get_1D_tiles(filepath, zoom + extra_zoom, higher_xpos)
            new_rows += old_rows

        # print("new_rows length", len(new_rows))
        to_return += [(tile_id, new_rows)]

    return to_return


def get_1D_tiles(db_file, zoom, tile_x_pos, num_tiles=1):
    """
    Retrieve a contiguous set of tiles from a db tile file.

    Parameters
    ----------
    db_file: str
        The filename of the sqlite db file
    zoom: int
        The zoom level
    tile_x_pos: int
        The position of the first tile
    num_tiles: int
        The number of tiles to retrieve

    Returns
    -------
    tiles: {pos: tile_value}
        A set of tiles, indexed by position
    """
    ts_info = tileset_info(db_file)
    version = ts_info["version"]

    conn = sqlite3.connect(db_file)

    c = conn.cursor()

    tile_width = ts_info["max_width"] / 2 ** zoom

    tile_start_pos = tile_width * tile_x_pos
    tile_end_pos = tile_start_pos + num_tiles * tile_width

    query = """
    SELECT startPos, endPos, chrOffset, importance, fields, uid
    FROM intervals,position_index
    WHERE
        intervals.id=position_index.id AND
        zoomLevel <= {} AND
        rEndPos >= {} AND
        rStartPos <= {}
    """.format(
        zoom, tile_start_pos, tile_end_pos
    )

    if version == 2:
        query = """
        SELECT startPos, endPos, chrOffset, importance, fields, uid
        FROM intervals,position_index
        WHERE
            intervals.id=position_index.id AND
            rStartZoomLevel <= {} AND
            rEndZoomLevel >= 0 AND
            rEndPos >= {} AND
            rStartPos <= {}
        """.format(
            zoom, tile_start_pos, tile_end_pos
        )

    if version == 3:
        query = """
        SELECT startPos, endPos, chrOffset, importance, fields, uid, name
        FROM intervals,position_index
        WHERE
            intervals.id=position_index.id AND
            rStartZoomLevel <= {} AND
            rEndZoomLevel >= 0 AND
            rEndPos >= {} AND
            rStartPos <= {}
        """.format(
            zoom, tile_start_pos, tile_end_pos
        )

    # import time
    # t1 = time.time()
    rows = c.execute(query).fetchall()
    # t2 = time.time()

    new_rows = []

    for r in rows:
        try:
            uid = r[5].decode("utf-8")
        except AttributeError:
            uid = r[5]

        x_start = r[0]
        x_end = r[1]

        for i in range(tile_x_pos, tile_x_pos + num_tiles):
            tile_x_start = i * tile_width
            tile_x_end = (i + 1) * tile_width

            if x_start < tile_x_end and x_end >= tile_x_start:
                to_add = {
                    "xStart": r[0],
                    "xEnd": r[1],
                    "chrOffset": r[2],
                    "importance": r[3],
                    "uid": uid,
                    "fields": r[4].split("\t"),
                }

                if version == 3:
                    to_add["name"] = r[6]

                new_rows += [to_add]
    conn.close()

    return new_rows


def list_items(db_file, start, end, max_entries=None):
    """
    List the entries between start and end

    Parameters
    ----------
    db_file: str
        The filename of the sqlite db file
    start_pos: int
        The start position from where to retrieve data
    end_pos: int
        The end position to get data
    max_entries: int
        The maximum number of results to return
    """
    ts_info = tileset_info(db_file)
    version = ts_info["version"]

    conn = sqlite3.connect(db_file)

    c = conn.cursor()

    # some large number because we want to extract all entries
    zoom = 100000

    query = """
    SELECT startPos, endPos, chrOffset, importance, fields, uid
    FROM intervals,position_index
    WHERE
        intervals.id=position_index.id AND
        zoomLevel <= {} AND
        rEndPos >= {} AND
        rStartPos <= {}
    """.format(
        zoom, start, end
    )

    if version == 2:
        query = """
        SELECT startPos, endPos, chrOffset, importance, fields, uid
        FROM intervals,position_index
        WHERE
            intervals.id=position_index.id AND
            rStartZoomLevel <= {} AND
            rEndZoomLevel >= 0 AND
            rEndPos >= {} AND
            rStartPos <= {}
        """.format(
            zoom, start, end
        )

    if version == 3:
        query = """
        SELECT startPos, endPos, chrOffset, importance, fields, uid, name
        FROM intervals,position_index
        WHERE
            intervals.id=position_index.id AND
            rStartZoomLevel <= {} AND
            rEndZoomLevel >= 0 AND
            rEndPos >= {} AND
            rStartPos <= {}
        """.format(
            zoom, start, end
        )
    if max_entries is not None:
        query += " LIMIT {}".format(max_entries)

    rows = c.execute(query).fetchall()

    new_rows = []

    for r in rows:
        try:
            uid = r[5].decode("utf-8")
        except AttributeError:
            uid = r[5]

        to_add = {
            "xStart": r[0],
            "xEnd": r[1],
            "chrOffset": r[2],
            "importance": r[3],
            "uid": uid,
            "fields": r[4].split("\t"),
        }

        if version == 3:
            to_add["name"] = r[6]

        new_rows += [to_add]
    conn.close()

    return new_rows



import clodius.tiles.bigbed as hgbb
import clodius.tiles.bigwig as hgbw
import os.path as op


def test_bigbed_tiles():
    filename = op.join(
        "data", "masterlist_DHSs_733samples_WM20180608_all_mean_signal_colorsMax.bed.bb"
    )

    tileset_info = hgbb.tileset_info(filename)
    num_chroms = len(tileset_info["chromsizes"])

    base_tile = hgbb.tiles(filename, ["x.0.0"])
    base_tile_data = base_tile[0][1]
    assert (num_chroms * hgbb.MAX_ELEMENTS) == len(
        base_tile_data
    ), "Number of chromosomes in bigBed file was {}\nExpected: {}\n".format(
        len(base_tile_data), (num_chroms * hgbb.MAX_ELEMENTS)
    )

    signif_tile = hgbb.tiles(filename, ["x.0.0.significant"])
    signif_tile_data = signif_tile[0][1]
    assert (num_chroms * hgbb.MAX_ELEMENTS) == len(
        signif_tile_data
    ), "Number of chromosomes in bigBed file was {}\nExpected: {}\n".format(
        len(signif_tile_data), (num_chroms * hgbb.MAX_ELEMENTS)
    )

    max_elements = [5, 10, 25, 100]
    for me in max_elements:
        max_elems_tile = hgbb.tiles(filename, ["x.0.0|max:{}".format(me)])
        max_elems_tile_data = max_elems_tile[0][1]
        for mtd in max_elems_tile_data:
            assert len(mtd) <= (
                me * hgbb.MAX_ELEMENTS
            ), "Number of observed elements with max element threshold was {}\nExpected: {}\n".format(
                len(mtd), (me * hgbb.MAX_ELEMENTS)
            )

    min_max_element_ranges = [(1, 5), (5, 10)]
    for mmer in min_max_element_ranges:
        min_element = mmer[0]
        max_element = mmer[1]
        min_max_elems_tile = hgbb.tiles(
            filename, ["x.0.0|min:{}|max:{}".format(min_element, max_element)]
        )
        min_max_elems_tile_data = min_max_elems_tile[0][1]
        mmtdl = len(min_max_elems_tile_data)
        assert (
            mmtdl >= min_element and mmtdl <= max_element * hgbb.MAX_ELEMENTS
        ), "Number of observed elements with min/max threshold was {}\nExpected: {} to {}\nAll data length: {}\nAll data: {}\n".format(
            mmtdl, min_element, max_element, mmtdl, min_max_elems_tile_data
        )


def test_tileset_info():
    filename = op.join(
        "data", "masterlist_DHSs_733samples_WM20180608_all_mean_signal_colorsMax.bed.bb"
    )

    tileset_info = hgbb.tileset_info(filename)

    assert len(tileset_info["range_modes"]) == 1
    assert tileset_info["range_modes"]["significant"]


def test_natsorted():
    chromname_tests = [
        {
            "input": ["2", "3", "4", "m", "x", "1", "y"],
            "expected": ["1", "2", "3", "4", "x", "y", "m"],
        },
        {
            "input": ["chr1", "chr4", "chr5", "chr2", "chr3", "chrMT", "chrY", "chrX"],
            "expected": [
                "chr1",
                "chr2",
                "chr3",
                "chr4",
                "chr5",
                "chrX",
                "chrY",
                "chrMT",
            ],
        },
    ]

    for test in chromname_tests:
        sorted_output = hgbw.natsorted(test["input"])
        assert (
            sorted_output == test["expected"]
        ), "Sorted output was %s\nExpected: %s" % (sorted_output, test["expected"])



import collections as col
import clodius.tiles.bedarcsdb as hgbad
import os.path as op


def get_counts(filename, zoom, pos):
    data = hgbad.tiles(filename, ["b.{}.{}".format(zoom, pos)])
    # print('data:', len(data[0][1][pos]))
    counts = col.defaultdict(int)

    for i, d in enumerate(data[0][1][pos]):
        chr1, chr2 = sorted([d["fields"][0], d["fields"][3]])
        # print('d:', d)

        counts[(chr1, chr2)] += 1
    return counts


def test_bedarcsdb_tiles1():
    filename = op.join(
        "data",
        "9ae0744a-9bc1-4cd7-b7cf-c6569ed9e4aa"
        ".pcawg_consensus_1.6.161022.somatic.sv.bedpe.multires.db",
    )

    tiles_to_get = [(2, 2)]

    for ttg in tiles_to_get:
        get_counts(filename, *ttg)
        # print(*ttg, counts)
    # counts1 = get_counts(filename, 0,0)
    # counts1 = get_counts(filename, 1,1)
    # counts2 = get_counts(filename, 2,2)
    # counts3 = get_counts(filename, 2,3)

    # print('1,1', counts1)
    # print('2,2', counts2)
    # print('2,3', counts3)

    # print('1', counts1[('21', 'X')])
    # print('2', counts1[('21', 'X')])

    # print("1,1", get_counts(filename, 1,1))
    # print('2,2', get_counts(filename, 2,2))

    """
    data = hgbad.tiles(filename, ["b.0.0"])
    # print('data:', len(data[0][1][0]))
    counts = col.defaultdict(int)

    for i,d in enumerate(data[0][1][0]):
        chr1,chr2 = sorted([d['fields'][0], d['fields'][3]])
        # print('d:', d)

        counts[(chr1,chr2)] += 1
    # print('counts:', counts)

    # print('counts:', counts)
    return

    data = hgbad.tiles(filename, ["b.4.11".format(pos)])
    # print('data:', len(data[0][1][11]))
    counts = col.defaultdict(int)

    for d in data[0][1][11]:
        chr1,chr2 = sorted([d['fields'][0], d['fields'][3]])

        counts[(chr1,chr2)] += 1
    # print('counts:', counts)

    data = hgbad.tiles(filename, ["b.4.10".format(pos)])
    # print('data:', len(data[0][1][10]))
    counts = col.defaultdict(int)

    for i,d in enumerate(data[0][1][10]):
        chr1,chr2 = sorted([d['fields'][0], d['fields'][3]])
        # print('d:', d)

        counts[(chr1,chr2)] += 1
    # print('counts:', counts)
    """


def test_bedarcsdb_tiles():
    """
    Retrieve a 1D tile from a 2d file
    """
    filename = op.join("data", "arrowhead_domains_short.txt.multires.db")

    pos = 2
    hgbad.tiles(filename, ["b.2.{}".format(pos)])

    # TODO: Do something with the data we get back?



from __future__ import division, print_function
import json

import numpy as np
import pandas as pd
import cooler
import h5py
import time

TILESIZE = 256

where = np.flatnonzero
chromsizes = cooler.read_chromsizes(
    "http://s3.amazonaws.com/pkerp/data/mm9/chromInfo.txt"
)  # defaults to reading chr#,X,Y,M
chromosomes = list(chromsizes.keys())
chromid_map = dict(zip(chromosomes, range(len(chromosomes))))
cumul_lengths = np.r_[0, np.cumsum(chromsizes)]


def absCoord2bin(c, pos):
    try:
        cid = where(cumul_lengths > pos)[0] - 1
    except IndexError:
        return c.info["nbins"]
    chrom = chromosomes[cid]
    relPos = pos - cumul_lengths[cid]
    return c.offset((chrom, relPos, chromsizes[chrom]))


def getData(FILEPATH, zoomLevel, startPos1, endPos1, startPos2, endPos2):

    groupname = str(zoomLevel)

    with h5py.File(FILEPATH, "r") as f:
        c = cooler.Cooler(f[groupname])

        i0 = absCoord2bin(c, startPos1)
        i1 = absCoord2bin(c, endPos1)
        j0 = absCoord2bin(c, startPos2)
        j1 = absCoord2bin(c, endPos2)
        mat = c.matrix(balance=True)[i0:i1, j0:j1]

    flat = list(mat.toarray().ravel())

    return json.dumps({"dense": flat})


def getData2(cooler_matrix, zoomLevel, startPos1, endPos1, startPos2, endPos2):

    c = cooler_matrix["cooler"]
    matrix = cooler_matrix["matrix"]

    i0 = absCoord2bin(c, startPos1)
    i1 = absCoord2bin(c, endPos1)
    j0 = absCoord2bin(c, startPos2)
    j1 = absCoord2bin(c, endPos2)

    if (i1 - i0) == 0 or (j1 - j0) == 0:
        return pd.DataFrame(columns=["genome_start", "genome_end", "balanced"])

    t1 = time.time()
    pixels = matrix[i0:i1, j0:j1]
    cid1 = pixels["chrom1"].apply(chromid_map.get)
    cid2 = pixels["chrom2"].apply(chromid_map.get)
    pixels["genome_start"] = cumul_lengths[cid1] + pixels["start1"]
    pixels["genome_end"] = cumul_lengths[cid2] + pixels["end2"]

    print("z:", zoomLevel, "t:", time.time() - t1)
    return pixels[["genome_start", "genome_end", "balanced"]]


def getData3(cooler_matrix, zoomLevel, startPos1, endPos1, startPos2, endPos2):
    c = cooler_matrix["cooler"]

    i0 = absCoord2bin(c, startPos1)
    i1 = absCoord2bin(c, endPos1)
    j0 = absCoord2bin(c, startPos2)
    j1 = absCoord2bin(c, endPos2)

    if (i1 - i0) == 0 or (j1 - j0) == 0:
        return pd.DataFrame(columns=["genome_start", "genome_end", "balanced"])

    pixels = c.matrix(as_pixels=True, max_chunk=np.inf)[i0:i1, j0:j1]

    if not len(pixels):
        return pd.DataFrame(columns=["genome_start", "genome_end", "balanced"])

    lo = min(i0, j0)
    hi = max(i1, j1)
    bins = c.bins()[["chrom", "start", "end", "weight"]][lo:hi]
    bins["chrom"] = bins["chrom"].cat.codes
    pixels = cooler.annotate(pixels, bins)
    pixels["genome_start"] = cumul_lengths[pixels["chrom1"]] + pixels["start1"]
    pixels["genome_end"] = cumul_lengths[pixels["chrom2"]] + pixels["end2"]
    pixels["balanced"] = pixels["count"] * pixels["weight1"] * pixels["weight2"]

    return pixels[["genome_start", "genome_end", "balanced"]]


def getInfo(FILEPATH):

    with h5py.File(FILEPATH, "r") as f:
        total_length = int(cumul_lengths[-1])
        binsize = int(f["0"].attrs["bin-size"])
        binsize = 1000
        n_tiles = total_length / binsize / TILESIZE
        print("total_length:", total_length, binsize, TILESIZE)
        n_zooms = int(np.ceil(np.log2(n_tiles)))
        max_width = binsize * TILESIZE * 2 ** n_zooms

        info = {
            "min_pos": [0.0, 0.0],
            "max_pos": [total_length, total_length],
            "max_zoom": n_zooms,
            "max_width": max_width,
            "bins_per_dimension": TILESIZE,
        }

    return info

    return json.dumps(info)



import collections as col
import cooler
import clodius.tiles.format as hgfo
import clodius.tiles.utils as hgut
import h5py
import itertools as it
import numpy as np
import pandas as pd
import logging

logger = logging.getLogger(__name__)

global mats
mats = {}

transform_descriptions = {}
transform_descriptions["weight"] = {"name": "ICE", "value": "weight"}
transform_descriptions["KR"] = {"name": "KR", "value": "KR"}
transform_descriptions["VC"] = {"name": "VC", "value": "VC"}
transform_descriptions["VC_SQRT"] = {"name": "VC_SQRT", "value": "VC_SQRT"}

TILE_SIZE = 256


def abs_coord_2_bin(c, abs_pos, chroms, chrom_cum_lengths, chrom_sizes):
    """Get bin ID from absolute coordinates.

    Args:
        c (Cooler): Cooler instance of a .cool file.
        abs_pos (int): Absolute coordinate to be translated.

    Returns:
        int: Bin number.
    """

    try:
        chr_id = np.flatnonzero(chrom_cum_lengths > abs_pos)[0] - 1
    except IndexError:
        return c.info["nbins"]

    chrom = chroms[chr_id]
    rel_pos = abs_pos - chrom_cum_lengths[chr_id]

    return c.offset((chrom, rel_pos, chrom_sizes[chrom]))


def get_chromosome_names_cumul_lengths(c):
    """
    Get the chromosome names and cumulative lengths:

    Args:

    c (Cooler): A cooler file

    Return:

    (names, sizes, lengths) -> (list(string), dict, np.array(int))
    """
    chrom_names = c.chromnames
    chrom_sizes = dict(c.chromsizes)
    chrom_cum_lengths = np.r_[0, np.cumsum(c.chromsizes.values)]
    return chrom_names, chrom_sizes, chrom_cum_lengths


def get_data(
    f,
    start_pos_1,
    end_pos_1,
    start_pos_2,
    end_pos_2,
    transform="default",
    resolution=None,
):
    """Get balanced pixel data.

    Args:
        f: h5py.File
            An HDF5 Group that contains the cooler for this resolution
        start_pos_1 (int): Test.
        end_pos_1 (int): Test.
        start_pos_2 (int): Test.
        end_pos_2 (int): Test.

    Returns:
        DataFrame: Annotated cooler pixels.
    """

    c = cooler.Cooler(f)

    (chroms, chrom_sizes, chrom_cum_lengths) = get_chromosome_names_cumul_lengths(c)

    i0 = abs_coord_2_bin(c, start_pos_1, chroms, chrom_cum_lengths, chrom_sizes)
    i1 = abs_coord_2_bin(c, end_pos_1, chroms, chrom_cum_lengths, chrom_sizes)

    j0 = abs_coord_2_bin(c, start_pos_2, chroms, chrom_cum_lengths, chrom_sizes)
    j1 = abs_coord_2_bin(c, end_pos_2, chroms, chrom_cum_lengths, chrom_sizes)

    matrix = c.matrix(as_pixels=True, balance=False, max_chunk=np.inf)

    if i0 >= matrix.shape[0] or j0 >= matrix.shape[1]:
        # query beyond the bounds of the matrix
        # return an empty matrix
        i0, i1, j0, j1 = 0, 0, 0, 0

        return (
            pd.DataFrame(columns=["genome_start1", "genome_start2", "balanced"]),
            (
                pd.DataFrame({"genome_start": [], "genome_end": [], "weight": []}),
                pd.DataFrame({"genome_start": [], "genome_end": [], "weight": []}),
            ),
        )
    else:
        # limit the range of the query to be within bounds
        i1 = min(i1, matrix.shape[0] - 1)
        j1 = min(j1, matrix.shape[1] - 1)

    pixels = matrix[i0 : i1 + 1, j0 : j1 + 1]

    """
    if not len(pixels):
        return (pd.DataFrame(columns=['genome_start1', 'genome_start2', 'balanced']), (None, None))
    """

    # select bin columns to extract
    cols = ["chrom", "start", "end"]
    if (transform == "default" and "weight" in c.bins()) or transform == "weight":
        cols.append("weight")
    elif transform in ("KR", "VC", "VC_SQRT"):
        cols.append(transform)

    bins = c.bins(convert_enum=False)[cols]
    pixels = cooler.annotate(pixels, bins)

    pixels["genome_start1"] = chrom_cum_lengths[pixels["chrom1"]] + pixels["start1"]
    pixels["genome_start2"] = chrom_cum_lengths[pixels["chrom2"]] + pixels["start2"]

    bins1 = bins[i0 : i1 + 1]
    bins2 = bins[j0 : j1 + 1]

    bins1["genome_start"] = chrom_cum_lengths[bins1["chrom"]] + bins1["start"]
    bins2["genome_start"] = chrom_cum_lengths[bins2["chrom"]] + bins2["start"]

    bins1["genome_end"] = chrom_cum_lengths[bins1["chrom"]] + bins1["end"]
    bins2["genome_end"] = chrom_cum_lengths[bins2["chrom"]] + bins2["end"]

    # apply transform
    if (transform == "default" and "weight" in c.bins()) or transform == "weight":
        pixels["balanced"] = pixels["count"] * pixels["weight1"] * pixels["weight2"]

        return (pixels[["genome_start1", "genome_start2", "balanced"]], (bins1, bins2))
    elif transform in ("KR", "VC", "VC_SQRT"):
        pixels["balanced"] = (
            pixels["count"] / pixels[transform + "1"] / pixels[transform + "2"]
        )

        bins1["weight"] = bins1[transform]
        bins2["weight"] = bins2[transform]

        return (pixels[["genome_start1", "genome_start2", "balanced"]], (bins1, bins2))
    else:
        return (pixels[["genome_start1", "genome_start2", "count"]], (None, None))


def _get_info_multi_v1(file_path):
    """Get information of a cooler file.

    Args:
        file_path (str): Path to a cooler file.

    Returns:
        dict: Dictionary containing basic information about the cooler file.
    """

    with h5py.File(file_path, "r") as f:
        max_zoom = f.attrs.get("max-zoom")

        if max_zoom is None:
            logger.info("no zoom found")
            raise ValueError("The `max_zoom` attribute is missing.")

        c = cooler.Cooler(f["0"])

        (chroms, chrom_sizes, chrom_cum_lengths) = get_chromosome_names_cumul_lengths(c)

        total_length = int(chrom_cum_lengths[-1])
        max_zoom = f.attrs["max-zoom"]
        bin_size = int(f[str(max_zoom)].attrs["bin-size"])

        max_width = bin_size * TILE_SIZE * 2 ** max_zoom

        # the list of available data transforms
        transforms = {}

        for i in range(max_zoom):
            f_for_zoom = f[str(i)]["bins"]

            if "weight" in f_for_zoom:
                transforms["weight"] = {"name": "ICE", "value": "weight"}
            if "KR" in f_for_zoom:
                transforms["KR"] = {"name": "KR", "value": "KR"}
            if "VC" in f_for_zoom:
                transforms["VC"] = {"name": "VC", "value": "VC"}
            if "VC_SQRT" in f_for_zoom:
                transforms["VC_SQRT"] = {"name": "VC_SQRT", "value": "VC_SQRT"}

        info = {
            "min_pos": [0.0, 0.0],
            "max_pos": [total_length, total_length],
            "max_zoom": max_zoom,
            "max_width": max_width,
            "bins_per_dimension": TILE_SIZE,
            "transforms": transforms.values(),
        }

    return info


def get_quadtree_depth(chromsizes, binsize):
    """
    Depth of quad tree necessary to tesselate the concatenated genome with quad
    tiles such that linear dimension of the tiles is a preset multiple of the
    genomic resolution.

    """
    tile_size_bp = TILE_SIZE * binsize
    min_tile_cover = np.ceil(sum(chromsizes) / tile_size_bp)
    return int(np.ceil(np.log2(min_tile_cover)))


def get_zoom_resolutions(chromsizes, base_res):
    return [
        base_res * 2 ** x for x in range(get_quadtree_depth(chromsizes, base_res) + 1)
    ]


def print_zoom_resolutions(chromsizes_file, base_res):
    """
    Print comma-separated list of zoom resolutions for a given genome
    and base resolution.

    """
    chromsizes = cooler.util.read_chromsizes(chromsizes_file, all_names=True)
    resolutions = get_zoom_resolutions(chromsizes, base_res)
    print(",".join(str(res) for res in resolutions))


def make_tiles(
    hdf_for_resolution,
    resolution,
    x_pos,
    y_pos,
    transform_type="default",
    x_width=1,
    y_width=1,
):
    """
    Generate tiles for a given location. This function retrieves tiles for
    a rectangular region of width x_width and height y_width

    Parameters
    ---------
    hdf_for_resolution: h5py.File
        An HDF group containing the cooler for the given resolution
    x_pos: int
        The starting x position
    y_pos: int
        The starting y position
    cooler_file: string
        The filename of the cooler file to get the data from
    x_width: int
        The number of tiles to retrieve along the x dimension
    y_width: int
        The number of tiles to retrieve along the y dimension

    Returns
    -------
    data_by_tilepos: {(x_pos, y_pos) : np.array}
        A dictionary of tile data indexed by tile positions
    """
    BINS_PER_TILE = 256

    tile_size = resolution * BINS_PER_TILE

    start1 = x_pos * tile_size
    end1 = (x_pos + x_width) * tile_size
    start2 = y_pos * tile_size
    end2 = (y_pos + y_width) * tile_size

    # print("resolution:", resolution)
    # print("tile_size:", tile_size)
    # print("transform_type:", transform_type);
    # print('start1:', start1, end1)
    # print('start2:', start2, end2)

    c = cooler.Cooler(hdf_for_resolution)
    (chroms, chrom_sizes, chrom_cum_lengths) = get_chromosome_names_cumul_lengths(c)

    total_length = sum(chrom_sizes.values())

    (data, (bins1, bins2)) = get_data(
        hdf_for_resolution,
        start1,
        end1 - 1,
        start2,
        end2 - 1,
        transform_type,
        resolution=resolution,
    )

    # print('start1', start1, 'end1', end1, 'weight', len(weight1), 'end1 - start1 / tile_size', (end1 - start1) / resolution)

    # print("data:", data)

    # print("x_width:", x_width)
    # print("y_width:", y_width)
    # split out the individual tiles
    data_by_tilepos = {}

    for x_offset in range(0, x_width):
        for y_offset in range(0, y_width):

            start1 = (x_pos + x_offset) * tile_size
            end1 = (x_pos + x_offset + 1) * tile_size
            start2 = (y_pos + y_offset) * tile_size
            end2 = (y_pos + y_offset + 1) * tile_size

            # print("resolution:", resolution)
            # print("tile_size", tile_size)
            # print("x_pos:", x_pos, "x_offset", x_offset)
            # print("start1", start1, 'end1', end1)
            # print("start2", start2, 'end2', end2)

            df = data[data["genome_start1"] >= start1]
            df = df[df["genome_start1"] < end1]

            df = df[df["genome_start2"] >= start2]
            df = df[df["genome_start2"] < end2]

            binsize = resolution

            j = ((df["genome_start1"].values - start1) // binsize).astype(int)
            i = ((df["genome_start2"].values - start2) // binsize).astype(int)

            if "balanced" in df:
                v = np.nan_to_num(df["balanced"].values)
            else:
                v = np.nan_to_num(df["count"].values)

            out = np.zeros((256, 256), dtype=np.float32)
            out[i, j] = v

            if bins1 is not None and bins2 is not None:
                sub_bins1 = bins1[bins1["genome_start"] >= start1]
                sub_bins2 = bins2[bins2["genome_start"] >= start2]

                sub_bins1 = sub_bins1[sub_bins1["genome_start"] < end1]
                sub_bins2 = sub_bins2[sub_bins2["genome_start"] < end2]

                # print("sub_bins1:", sub_bins1)

                nan_bins1 = sub_bins1[np.isnan(sub_bins1["weight"])]
                nan_bins2 = sub_bins2[np.isnan(sub_bins2["weight"])]

                bi = ((nan_bins1["genome_start"].values - start1) // binsize).astype(
                    int
                )
                bj = ((nan_bins2["genome_start"].values - start2) // binsize).astype(
                    int
                )

                bend1 = (
                    (np.array(range(total_length, int(end1), int(resolution))) - start1)
                    // binsize
                ).astype(int)
                bend2 = (
                    (np.array(range(total_length, int(end2), int(resolution))) - start2)
                    // binsize
                ).astype(int)

                bend1 = bend1[bend1 >= 0]
                bend2 = bend2[bend2 >= 0]

                out[:, bi] = np.nan
                out[bj, :] = np.nan

                out[:, bend1] = np.nan
                out[bend2, :] = np.nan

            # print('sum(isnan1)', isnan1-1)
            # print('out.ravel()', sum(np.isnan(out.ravel())), len(out.ravel()))
            data_by_tilepos[(x_pos + x_offset, y_pos + y_offset)] = out.ravel()

    return data_by_tilepos


def bin_tiles_by_zoom_level_and_transform(tile_ids):
    """
    Place these tiles into separate lists according to their
    zoom level and transform type

    Parameters
    ----------
    tile_ids: [str,...]
        A list of tile_ids (e.g. xyx.0.0.1) identifying the tiles
        to be retrieved

    Returns
    -------
    tile_lists: {(zoomLevel, transformType): [tile_id, tile_id]}
        A dictionary of tile ids
    """
    tile_id_lists = col.defaultdict(set)

    for tile_id in tile_ids:
        tile_id_parts = tile_id.split(".")
        tile_position = list(map(int, tile_id_parts[1:4]))
        zoom_level = tile_position[0]

        transform_method = get_transform_type(tile_id)

        tile_id_lists[(zoom_level, transform_method)].add(tile_id)

    return tile_id_lists


def get_transform_type(tile_id):
    """
    Get the transform type specified in the tile id.
    Parameters
    ----------
    cooler_tile_id: str
        A tile id for a 2D tile (cooler)
    Returns
    -------
    transform_type: str
        The transform type requested for this tile
    """
    tile_id_parts = tile_id.split(".")

    if len(tile_id_parts) > 4:
        transform_method = tile_id_parts[4]
    else:
        transform_method = "default"

    return transform_method


def get_available_transforms(cooler):
    """
    Get the available resolutions from a single cooler file.
    Parameters
    ----------
    cooler: h5py File
        A cooler file containing binned 2D data
    Returns
    -------
    transforms: dict
        A list of transforms available for this dataset
    """
    transforms = set()

    f_for_zoom = cooler["bins"]

    if "weight" in f_for_zoom:
        transforms.add("weight")
    if "KR" in f_for_zoom:
        transforms.add("KR")
    if "VC" in f_for_zoom:
        transforms.add("VC")
    if "VC_SQRT" in f_for_zoom:
        transforms.add("VC_SQRT")

    return transforms


def make_mats(filepath):
    """
    Create the file handle and tileset info for a cooler
    tileset
    """
    f = h5py.File(filepath, "r")

    if "resolutions" in f:
        # this file contains raw resolutions so it'll return a different
        # sort of tileset info
        info = {"resolutions": tuple(sorted(map(int, list(f["resolutions"].keys()))))}
        mats[filepath] = [f, info]

        # see which transforms are available, a transform has to be
        # available at every available resolution in order for it to
        # be provided as an option
        available_transforms_per_resolution = {}

        for resolution in info["resolutions"]:
            available_transforms_per_resolution[resolution] = get_available_transforms(
                f["resolutions"][str(resolution)]
            )

        all_available_transforms = set.intersection(
            *available_transforms_per_resolution.values()
        )

        info["transforms"] = [
            transform_descriptions[t] for t in all_available_transforms
        ]

        # get the genome size
        resolution = list(f["resolutions"].keys())[0]
        genome_length = int(sum(f["resolutions"][resolution]["chroms"]["length"]))

        info["max_pos"] = [genome_length, genome_length]
        info["min_pos"] = [1, 1]

        c = cooler.Cooler(f["resolutions"][resolution])
        info["chromsizes"] = [[x[0], int(x[1])] for x in c.chromsizes.iteritems()]
        if "storage-mode" in c.info and c.info["storage-mode"] == "square":
            info["mirror_tiles"] = "false"
    else:
        info = _get_info_multi_v1(filepath)

        c = cooler.Cooler(f["0"])

        info["chromsizes"] = [[x[0], int(x[1])] for x in c.chromsizes.iteritems()]
        info["min_pos"] = [int(m) for m in info["min_pos"]]
        info["max_pos"] = [int(m) for m in info["max_pos"]]
        info["max_zoom"] = int(info["max_zoom"])
        info["max_width"] = int(info["max_width"])

        if "transforms" in info:
            info["transforms"] = list(info["transforms"])

        # legacy metadata for non-symmetric matrices
        if "symmetric" in c.info and not c.info["symmetric"]:
            info["mirror_tiles"] = "false"
        if "storage-mode" in c.info and c.info["storage-mode"] == "square":
            info["mirror_tiles"] = "false"

        mats[filepath] = [f, info]

    return f, info


def tileset_info(filepath):
    """
    Get the tileset info for a cooler file

    Parameters:
    -----------

    filepath: str
        The location of the cooler file
    """
    if filepath in mats:
        return mats[filepath][1]
    else:
        (f, info) = make_mats(filepath)

        return info


def add_transform_type(tile_id):
    """
    Add a transform type to a cooler tile id if it's not already
    present.

    Parameters
    ----------
    tile_id: str
        A tile id (e.g. xyz.0.1.0)

    Returns
    -------
    new_tile_id: str
        A formatted tile id, potentially with an added transform_type
    """
    tile_id_parts = tile_id.split(".")
    tileset_uuid = tile_id_parts[0]
    tile_position = tile_id_parts[1:4]

    transform_type = get_transform_type(tile_id)
    new_tile_id = ".".join([tileset_uuid] + tile_position + [transform_type])
    return new_tile_id


def tiles(filepath, tile_ids):
    """
    """
    transform_id_to_original_id = {}

    new_tile_ids = []

    for tile_id in tile_ids:
        new_tile_id = add_transform_type(tile_id)
        transform_id_to_original_id[new_tile_id] = tile_id
        new_tile_ids += [new_tile_id]

    generated_tiles = generate_tiles(filepath, new_tile_ids)

    tiles_to_return = []
    for tile_id, tile_value in generated_tiles:
        if tile_id in transform_id_to_original_id:
            original_tile_id = transform_id_to_original_id[tile_id]

            tiles_to_return += [(original_tile_id, tile_value)]

    return tiles_to_return


def generate_tiles(filepath, tile_ids):
    """
    Generate tiles from a cooler file.
    Parameters
    ----------
    tileset: tilesets.models.Tileset object
        The tileset that the tile ids should be retrieved from
    tile_ids: [str,...]
        A list of tile_ids (e.g. xyx.0.0.1) identifying the tiles
        to be retrieved
    Returns
    -------
    generated_tiles: [(tile_id, tile_data),...]
        A list of tile_id, tile_data tuples
    """
    BINS_PER_TILE = 256

    if filepath not in mats:
        # check if this tileset is open
        make_mats(filepath)

    tileset_file_and_info = mats[filepath]

    tile_ids_by_zoom_and_transform = bin_tiles_by_zoom_level_and_transform(
        tile_ids
    ).values()
    partitioned_tile_ids = list(
        it.chain(
            *[
                hgut.partition_by_adjacent_tiles(t)
                for t in tile_ids_by_zoom_and_transform
            ]
        )
    )

    generated_tiles = []

    for tile_group in partitioned_tile_ids:
        zoom_level = int(tile_group[0].split(".")[1])
        tileset_id = tile_group[0].split(".")[0]
        transform_type = get_transform_type(tile_group[0])
        tileset_info = tileset_file_and_info[1]
        tileset_file = tileset_file_and_info[0]

        if "resolutions" in tileset_info:
            sorted_resolutions = sorted(
                [int(r) for r in tileset_info["resolutions"]], reverse=True
            )
            if zoom_level > len(sorted_resolutions):
                # this tile has too high of a zoom level specified
                continue

            resolution = sorted_resolutions[zoom_level]
            hdf_for_resolution = tileset_file["resolutions"][str(resolution)]
        else:
            if zoom_level > tileset_info["max_zoom"]:
                # this tile has too high of a zoom level specified
                continue
            hdf_for_resolution = tileset_file[str(zoom_level)]
            resolution = (tileset_info["max_width"] / 2 ** zoom_level) / BINS_PER_TILE

        tile_positions = [[int(x) for x in t.split(".")[2:4]] for t in tile_group]

        # filter for tiles that are in bounds for this zoom level
        tile_positions = list(
            filter(lambda x: x[0] < tileset_info["max_pos"][0] + 1, tile_positions)
        )
        tile_positions = list(
            filter(lambda x: x[1] < tileset_info["max_pos"][1] + 1, tile_positions)
        )

        if len(tile_positions) == 0:
            # no in bounds tiles
            continue

        minx = min([t[0] for t in tile_positions])
        maxx = max([t[0] for t in tile_positions])

        miny = min([t[1] for t in tile_positions])
        maxy = max([t[1] for t in tile_positions])

        tile_data_by_position = make_tiles(
            hdf_for_resolution,
            resolution,
            minx,
            miny,
            transform_type,
            maxx - minx + 1,
            maxy - miny + 1,
        )

        tiles = [
            (
                ".".join(
                    map(
                        str,
                        [tileset_id] + [zoom_level] + list(position) + [transform_type],
                    )
                ),
                hgfo.format_dense_tile(tile_data),
            )
            for (position, tile_data) in tile_data_by_position.items()
        ]

        generated_tiles += tiles

    return generated_tiles



import collections as col
import functools as ft
import glob
import itertools as it
import os.path as op


class ParallelData:
    def __init__(self, data):
        self.data = data

    def map(self, func):
        return ParallelData([func(d) for d in self.data])

    def count(self):
        return len(self.data)

    def union(self, other_data):
        self.data = self.data + other_data.data
        return self

    def getNumPartitions(self):
        return 1

    def countByKey(self):
        """
        Count the number of elements with a particular key.
        """
        counts = col.defaultdict(int)

        for d in self.data:
            counts[d[0]] += 1

        return counts

    def coalesce(self, num):
        return self

    def groupByKey(self):
        """
        When called on a dataset of (K, V) pairs, returns a dataset of (K, Iterable<V>) pairs.
        """
        buckets = col.defaultdict(list)

        for d in self.data:
            buckets[d[0]].append(d[1])

        return ParallelData(buckets.items())

    def flatMap(self, func):
        """
        Flatten a list of results mapped to the data
        and return as one large list.
        """
        result = map(func, self.data)

        return ParallelData(list(it.chain.from_iterable(result)))

    def foreach(self, func):
        for d in self.data:
            func(d)

    def foreachPartition(self, func):
        func(self.data)

    def reduceByKey(self, func):
        """
        When called on a dataset of (K, V) pairs, returns a dataset of (K, V)
        pairs where the values for each key are aggregated using the given
        reduce function func, which must be of type (V,V) => V. Like in
        groupByKey, the number of reduce tasks is configurable through an
        optional second argument.
        """
        buckets = col.defaultdict(list)
        for d in self.data:
            buckets[d[0]].append(d[1])

        reduced_buckets = dict()
        for key in buckets:
            reduced_buckets[key] = ft.reduce(func, buckets[key])

        return ParallelData(reduced_buckets.items())

    def aggregateByKey(self, start_val, seq_func, comb_func):
        buckets = col.defaultdict(list)
        for d in self.data:
            buckets[d[0]].append(d[1])

        reduced_buckets = dict()
        for key in buckets:
            comb_val = start_val.copy()
            for val in buckets[key]:
                comb_val = seq_func(comb_val, val)
            reduced_buckets[key] = comb_val

        return ParallelData(reduced_buckets.items())

    def reduce(self, func):
        return ft.reduce(func, self.data)

    def collect(self):
        return self.data

    def take(self, n):
        return self.data[:n]


class FakeSparkContext:
    """
    Emulate a SparkContext for local processing.
    """

    def __init__(self):
        pass

    @staticmethod
    def parallelize(data):
        return ParallelData(data)

    @staticmethod
    def singleTextFile(filename):
        """
        Load a single file as a ParallelData set.

        :param filename: A file to be loaded line by line
        @return: A ParallelData object wrapping the lines in the file.
        """
        with open(filename, "r") as f:
            return ParallelData(list(map(lambda x: x.strip(), f.readlines())))

    @staticmethod
    def textFile(filename):
        """
        Load a filename as a text file. Filename can be either a single file
        or a directory containing a multitude of 'part-*' files.

        :param filename: The name of the file (or directory) containing the data which
                         we want to load one by one
        :return: A ParallelData object containing all of the lines of the file or files
        """
        if op.isdir(filename):
            parts_files = glob.glob(op.join(filename, "part-*"))
            p = FakeSparkContext.parallelize([])

            for filename in parts_files:
                p = p.union(FakeSparkContext.singleTextFile(filename))
            return p
        else:
            return FakeSparkContext.singleTextFile(filename)



# -*- coding: utf-8 -*-
from __future__ import division, print_function
import click


# Monkey patch
click.core._verify_python3_env = lambda: None


CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    pass


from . import (  # noqa: F401
    # Another approach would be __all__:
    # https://docs.python.org/2/tutorial/modules.html#importing-from-a-package
    aggregate,
    convert,
)



import base64
import os
import sqlite3


def get_tileset_info(tileset):
    if not os.path.isfile(tileset):
        return {"error": "Tileset info is not available!"}

    db = sqlite3.connect(tileset)

    res = db.execute("SELECT * FROM tileset_info").fetchone()

    o = {"tile_size": res[5], "max_zoom": res[6], "max_size": res[7]}

    try:
        o["width"] = res[8]
        o["height"] = res[9]
    except IndexError:
        pass

    try:
        o["dtype"] = res[10]
    except IndexError:
        pass

    return o


def get_tiles(filename, tile_ids, raw):
    """
    Generate tiles from a imtiles file.
    Parameters
    ----------
    tileset: tilesets.models.Tileset object
        The tileset that the tile ids should be retrieved from
    tile_ids: [str,...]
        A list of tile_ids (e.g. xyx.0.0.1) identifying the tiles
        to be retrieved
    Returns
    -------
    generated_tiles: [(tile_id, tile_data),...]
        A list of tile_id, tile_data tuples
    """
    # Connect to SQLite db
    db = sqlite3.connect(filename)

    generate_tiles = []

    generate_image = raw and len(tile_ids)

    for tile_id in tile_ids:
        id = tile_id[tile_id.find(".") + 1 :].split(".")

        sql = "SELECT image FROM tiles WHERE z = :z AND y = :y AND x = :x"
        param = {"z": int(id[0]), "y": int(id[1]), "x": int(id[2])}
        res = db.execute(sql, param).fetchone()

        if res:
            image_blob = res[0]

            if generate_image:
                tile_data = {"image": image_blob}
            else:
                tile_data = {"dense": base64.b64encode(image_blob).decode("latin-1")}

            generate_tiles.append((tile_id, tile_data))

    return generate_tiles



import collections as col
import itertools as it
import sqlite3


def tiles(filepath, tile_ids):
    """
    Generate tiles from this dataset.

    Parameters
    ----------
    filepath: str
        The filename of the sqlite db file
    tile_ids: [str...]
        A list of tile ids of the form

    Returns
    -------
    tiles: [(tile_id, tile_value),...]
        A list of values indexed by the tile position
    """
    to_return = []

    for tile_id in tile_ids:
        parts = tile_id.split(".")
        zoom = int(parts[1])
        xpos = int(parts[2])

        extra_zoom = 3
        new_rows = {}
        new_rows[xpos] = []

        for j in range(2 ** extra_zoom):
            # the old rows are indexed by the higher
            # resolution tile numbers
            higher_xpos = 2 ** extra_zoom * xpos + j
            old_rows = get_1D_tiles(filepath, zoom + extra_zoom, higher_xpos)
            new_rows[xpos] += old_rows[higher_xpos]

        to_return += [(tile_id, new_rows)]

    return to_return


def get_1D_tiles(db_file, zoom, tile_x_pos, numx=1, numy=1):
    """
    Retrieve a contiguous set of tiles from a 2D db tile file.

    Parameters
    ----------
    db_file: str
        The filename of the sqlite db file
    zoom: int
        The zoom level
    tile_x_pos: int
        The x position of the first tile
    numx: int
        The width of the block of tiles to retrieve

    Returns
    -------
    tiles: {pos: tile_value}
        A set of tiles, indexed by position
    """
    tileset_info = get_2d_tileset_info(db_file)

    conn = sqlite3.connect(db_file)

    c = conn.cursor()
    tile_width = tileset_info["max_width"] / 2 ** zoom

    tile_x_start_pos = tile_width * tile_x_pos
    tile_x_end_pos = tile_x_start_pos + (numx * tile_width)

    query = """
    SELECT fromX, toX, fromY, toY, chrOffset, importance, fields, uid
    FROM intervals,position_index
    WHERE
        intervals.id=position_index.id AND
        zoomLevel <= {} AND
        rToX >= {} AND
        rFromX <= {} ORDER BY importance
    """.format(
        zoom, tile_x_start_pos, tile_x_end_pos
    )

    rows = c.execute(query).fetchall()

    query1 = """
    SELECT fromX, toX, fromY, toY, chrOffset, importance, fields, uid
    FROM intervals,position_index
    WHERE
        intervals.id=position_index.id AND
        zoomLevel <= {} AND
        rToY >= {} AND
        rFromY <= {} ORDER BY importance
    """.format(
        zoom, tile_x_start_pos, tile_x_end_pos
    )

    rows1 = c.execute(query1).fetchall()

    seen_uids = set()

    new_rows = col.defaultdict(list)
    all_rows = list(sorted(it.chain(rows, rows1), key=lambda x: -x[5]))
    # print("total_len", len(all_rows))

    MAX_ROWS = 25
    for r in all_rows:
        # fields = r[6].split('\t')
        # print('fields', sorted([fields[0], fields[3]]), r[5])
        try:
            uid = r[7].decode("utf-8")
        except AttributeError:
            uid = r[7]

        if uid in seen_uids:
            continue

        seen_uids.add(uid)
        if len(seen_uids) > MAX_ROWS:
            break

        x_start = r[0]
        x_end = r[1]
        y_start = r[2]
        y_end = r[3]

        for i in range(tile_x_pos, tile_x_pos + numx):
            tile_x_start = i * tile_width
            tile_x_end = (i + 1) * tile_width

            if (x_start < tile_x_end and x_end >= tile_x_start) or (
                y_start < tile_x_end and y_end >= tile_x_start
            ):
                # add the position offset to the returned values
                new_rows[i] += [
                    {
                        "xStart": r[0],
                        "xEnd": r[1],
                        "yStart": r[2],
                        "yEnd": r[3],
                        "chrOffset": r[4],
                        "importance": r[5],
                        "uid": uid,
                        "fields": r[6].split("\t"),
                    }
                ]
    conn.close()

    return new_rows


def get_2d_tileset_info(db_file):
    conn = sqlite3.connect(db_file)
    c = conn.cursor()

    row = c.execute("SELECT * from tileset_info").fetchone()
    tileset_info = {
        "zoom_step": row[0],
        "max_length": row[1],
        "assembly": row[2],
        "chrom_names": row[3],
        "chrom_sizes": row[4],
        "tile_size": row[5],
        "max_zoom": row[6],
        "max_width": row[7],
        "min_pos": [1, 1],
        "max_pos": [row[1], row[1]],
    }
    conn.close()

    return tileset_info



from setuptools import setup
import os
import io


HERE = os.path.dirname(os.path.abspath(__file__))


def read(*parts, **kwargs):
    filepath = os.path.join(HERE, *parts)
    encoding = kwargs.pop("encoding", "utf-8")
    with io.open(filepath, encoding=encoding) as fh:
        text = fh.read()
    return text


def get_requirements(path):
    content = read(path)
    return [req for req in content.split("\n") if req != "" and not req.startswith("#")]


setup_requires = ["numpy"]

install_requires = get_requirements("requirements.txt")

setup(
    name="clodius",
    version="0.14.0",
    description="Tile generation for big data",
    author="Peter Kerpedjiev",
    author_email="pkerpedjiev@gmail.com",
    url="",
    packages=["clodius", "clodius.cli", "clodius.tiles"],
    setup_requires=setup_requires,
    install_requires=install_requires,
    scripts=["scripts/tsv_to_mrmatrix.py"],
    entry_points={"console_scripts": ["clodius = clodius.cli.aggregate:cli"]},
)



#!/usr/bin/python

import csv
import dask.array as da
import h5py
import math
import numpy as np
import argparse
import time


def coarsen(f, type, tile_size=256):
    """
    Create data pyramid.
    """
    grid = f["resolutions"]["1"][type]
    top_n = grid.shape[0]
    max_zoom = math.ceil(math.log(top_n / tile_size) / math.log(2))

    chunk_size = tile_size * 16
    curr_size = grid.shape
    dask_dset = da.from_array(grid, chunks=(chunk_size, chunk_size))

    r = f["resolutions"]
    curr_resolution = 1

    while curr_resolution < 2 ** max_zoom:
        curr_size = tuple(np.array(curr_size) / 2)
        print("coarsening")
        curr_resolution *= 2

        print("curr_size:", curr_size)
        group_name = "{}{}".format(
            curr_resolution, "" if type == "values" else "-" + type
        )
        g = r.create_group(group_name)
        values = g.require_dataset(
            type, curr_size, dtype="f4", compression="lzf", fillvalue=np.nan
        )

        dask_dset = dask_dset.rechunk((chunk_size, chunk_size))
        dask_dset = da.coarsen(np.nansum, dask_dset, {0: 2, 1: 2})
        da.store(dask_dset, values)


def parse(input_handle, output_hdf5, height, width, delimiter, first_n, is_labelled):
    reader = csv.reader(input_handle, delimiter=delimiter)
    if is_labelled:
        first_row = next(reader)
        col_labels = first_row[1 : (first_n + 1) if first_n else None]
        output_hdf5.create_dataset(
            "col_labels",
            data=np.array(col_labels, dtype=h5py.special_dtype(vlen=str)),
            compression="lzf",
        )

    tile_size = 256
    limit = max(height, width)
    max_zoom = math.ceil(math.log(limit / tile_size) / math.log(2))
    max_dimension = tile_size * 2 ** max_zoom

    g = output_hdf5.create_group("resolutions")
    g1 = g.create_group("1")
    val_ds = g1.create_dataset(
        "values",
        (max_dimension, max_dimension),
        dtype="f4",
        compression="lzf",
        fillvalue=np.nan,
    )
    nan_ds = g1.create_dataset(
        "nan_values",
        (max_dimension, max_dimension),
        dtype="f4",
        compression="lzf",
        fillvalue=0,
    )

    start_time = time.time()
    counter = 0
    row_labels = []
    for row in reader:
        if is_labelled:
            offset = 1
            row_labels.append(row[0])
        else:
            offset = 0

        val_np = np.array([float(p) for p in row[offset:]])
        val_ds[counter, : len(val_np)] = val_np

        nan_np = np.array([0 for p in row[offset:]])
        nan_ds[counter, : len(nan_np)] = nan_np

        counter += 1
        if counter == first_n:
            break

        time_elapsed = time.time() - start_time
        time_per_entry = time_elapsed / counter

        time_remaining = time_per_entry * (height - counter)
        print(
            "counter:",
            counter,
            "sum(x):",
            sum(val_np),
            "time remaining: {:d} seconval_ds".format(int(time_remaining)),
        )

    if is_labelled:
        output_hdf5.create_dataset(
            "row_labels",
            data=np.array(row_labels, dtype=h5py.special_dtype(vlen=str)),
            compression="lzf",
        )
    coarsen(output_hdf5, "values")
    coarsen(output_hdf5, "nan_values")
    output_hdf5.close()


def get_height(input_path, is_labelled=True):
    """
    We need to scan the file once just to see how many lines it contains.
    If it is tall and narrow, the first tile will need to be larger than just
    looking at the width of the first row would suggest.
    """
    with open(input_path) as f:
        for i, l in enumerate(f):
            pass
    if is_labelled:
        return i
    else:
        return i + 1


def get_width(input_path, is_labelled, delimiter="\t"):
    """
    Assume the number of elements in the first row is the total width.
    """
    with open(input_path, "r", newline="") as input_handle:
        reader = csv.reader(input_handle, delimiter=delimiter)
        len_row = len(next(reader))
        if is_labelled:
            return len_row - 1
        return len_row


def main():
    parser = argparse.ArgumentParser(
        description="""
        Given a tab-delimited file, produces an HDF5 file with mrmatrix
        ("multi-resolution matrix") structure: Under the "resolutions"
        group are datasets, named with successive powers of 2,
        which represent successively higher aggregations of the input.
    """
    )
    parser.add_argument("input_file", help="TSV file path")
    parser.add_argument("output_file", help="HDF5 file")
    parser.add_argument(
        "-d",
        "--delimiter",
        type=str,
        default="\t",
        metavar="D",
        help="Delimiter; defaults to tab",
    )
    parser.add_argument(
        "-n",
        "--first-n",
        type=int,
        default=None,
        metavar="N",
        help="Only read first N columns from first N rows",
    )
    parser.add_argument(
        "-l",
        "--labelled",
        action="store_true",
        help="TSV Matrix has column and row labels",
    )
    args = parser.parse_args()

    height = get_height(args.input_file, is_labelled=args.labelled)
    width = get_width(
        args.input_file, is_labelled=args.labelled, delimiter=args.delimiter
    )
    print("height:", height)
    print("width:", width)

    f_in = open(args.input_file, "r", newline="")

    parse(
        f_in,
        h5py.File(args.output_file, "w"),
        height=height,
        width=width,
        delimiter=args.delimiter,
        first_n=args.first_n,
        is_labelled=args.labelled,
    )

    f = h5py.File(args.output_file, "r")
    print("sum1:", np.nansum(f["resolutions"]["1"]["values"][0]))


if __name__ == "__main__":
    main()



import os.path as op
import base64

import h5py
import pytest
import clodius.tiles.multivec as hgmu


def test_multivec():
    filename = op.join("test/sample_data", "sample_gwas.multires.mv5")
    with h5py.File(filename, "r") as h5:
        tile_size = h5["info"].attrs["tile-size"]
        resolutions = list(h5["resolutions"].keys())
        reso = resolutions[0]
        chroms = h5[f"resolutions/{reso}/chroms/name"][:]
        num_rows = h5[f"resolutions/{reso}/values"][chroms[0]].shape[1]
        total_length = sum(h5["chroms/length"])
    # info
    info = hgmu.tileset_info(filename)
    assert info["shape"] == [tile_size, num_rows]
    assert info["tile_size"] == tile_size
    assert info["max_pos"] == total_length
    assert set(info["resolutions"]) == set(int(reso) for reso in resolutions)

    # get_single_tile
    test_tile = hgmu.get_single_tile(filename, [0, 0])
    assert list(test_tile.shape)[::-1] == info["shape"]
    with pytest.raises(IndexError):
        hgmu.get_single_tile(filename, [len(resolutions), 0])

    # tiles
    tids = [f"test_uuid.{level}.0.1231.123" for level in range(len(resolutions))]
    tiles = hgmu.tiles(filename, tids)
    for tile_id, tile_value in tiles:
        tile_pos = [int(i) for i in tile_id.split(".")[1:3]]
        single_tile = hgmu.get_single_tile(filename, tile_pos).astype(
            tile_value["dtype"]
        )
        assert (
            base64.b64encode(single_tile.ravel()).decode("utf-8") == tile_value["dense"]
        )



import clodius.tiles.beddb as hgbe
import os.path as op


def test_get_tiles():
    filename = op.join("data", "corrected.geneListwithStrand.bed.multires")

    hgbe.tiles(filename, ["x.1.0", "x.1.1"])
    # TODO: Do something with the return value


def test_list_items():
    filename = op.join("data", "gene_annotations.short.db")

    items = hgbe.list_items(filename, 0, 100000000, max_entries=100)
    assert "name" not in items[0]
    # TODO: Do something with the return value


def test_name_in_tile():
    filename = op.join("data", "geneAnnotationsExonUnions.1000.bed.v3.beddb")

    tiles = hgbe.tiles(filename, ["x.1.0", "x.1.1"])

    assert "name" in tiles[0][1][0]



from __future__ import print_function

import clodius.db_tiles as cdt
import clodius.cli.aggregate as cca
import os
import os.path as op
import sys
import tempfile

sys.path.append("scripts")

testdir = op.realpath(op.dirname(__file__))


def test_clodius_aggregate_bedpe():
    input_file = op.join(testdir, "sample_data", "isidro.bedpe")

    with tempfile.TemporaryDirectory() as tmpdirname:
        output_file = op.join(tmpdirname, "isidro.bed2ddb")

        cca._bedpe(
            input_file,
            output_file,
            "b37",
            importance_column=None,
            chromosome=None,
            max_per_tile=100,
            tile_size=1024,
            has_header=True,
        )

        """
        runner = clt.CliRunner()
        result = runner.invoke(
                cca.bedpe,
                [input_file,
                '--output-file', output_file,
                '--importance-column', 'random',
                '--has-header',
                '--assembly', 'b37'])

        # print('output:', result.output, result)
        assert(result.exit_code == 0)
        """

        cdt.get_2d_tiles(output_file, 0, 0, 0)
        # print("entries:", entries)

        cdt.get_tileset_info(output_file)
        # TODO: Make assertions about result
        # print('tileset_info', tileset_info)

        cdt.get_2d_tiles(output_file, 1, 0, 0, numx=2, numy=2)
        # TODO: Make assertions about result
        # print("entries:", entries)

        cdt.get_tileset_info(output_file)
        # TODO: Make assertion


def test_clodius_aggregate_bedpe2():
    """Use galGal6 chromsizes file"""
    input_file = op.join(testdir, "sample_data", "galGal6.bed")
    chromsizes_file = op.join(testdir, "sample_data", "galGal6.chrom.sizes")
    expected_file = op.join(testdir, "sample_data", "galGal6.bed.multires.db")

    with tempfile.TemporaryDirectory() as tmpdirname:
        output_file = op.join(tmpdirname, "blah.bed2ddb")
        # the test is here to ensure that this doesn't raise an error
        cca._bedpe(
            input_file,
            output_file,
            None,
            chr1_col=1,
            chr2_col=1,
            from1_col=2,
            from2_col=2,
            to1_col=3,
            to2_col=3,
            importance_column=None,
            chromosome=None,
            chromsizes_filename=chromsizes_file,
            max_per_tile=100,
            tile_size=1024,
            has_header=True,
        )

        tsinfo = cdt.get_tileset_info(output_file)

        stat_output = os.stat(output_file)
        stat_expected = os.stat(expected_file)

        assert tsinfo["max_length"] == 1065365426
        assert stat_output.st_size == stat_expected.st_size
